"""Sphinx plugins for web API docs."""

import ast
import inspect
import json
import logging
import re
import sys

from beanbag_docutils.sphinx.ext.http_role import (
    DEFAULT_HTTP_STATUS_CODES_URL, HTTP_STATUS_CODES)
from django.contrib.auth.models import User
from django.http import HttpRequest
from django.template.defaultfilters import title
from django.utils import six
from djblets.features.testing import override_feature_checks
from djblets.util.http import is_mimetype_a
from djblets.webapi.resources import get_resource_from_class, WebAPIResource
from djblets.webapi.responses import WebAPIResponseError
from docutils import nodes
from docutils.parsers.rst import Directive, DirectiveError, directives
from docutils.statemachine import StringList, ViewList, string2lines
from reviewboard import initialize
from reviewboard.webapi.resources import resources
from sphinx import addnodes
from sphinx.util import docname_join
from sphinx.util.docstrings import prepare_docstring


# Mapping of mimetypes to language names for syntax highlighting.
MIMETYPE_LANGUAGES = [
    ('application/json', 'javascript'),
    ('application/xml', 'xml'),
    ('text/x-patch', 'diff'),
]


# Initialize Review Board
initialize()


# Build the list of parents.
resources.root.get_url_patterns()


class ResourceNotFound(Exception):
    def __init__(self, directive, classname):
        self.classname = classname
        self.error_node = [
            directive.state_machine.reporter.error(
                str(self),
                line=directive.lineno)
        ]

    def __str__(self):
        return ('Unable to import the web API resource class "%s"'
                % self.classname)


class ErrorNotFound(Exception):
    def __init__(self, directive, classname):
        self.error_node = [
            directive.state_machine.reporter.error(
                'Unable to import the web API error class "%s"' % classname,
                line=directive.lineno)
        ]


class DummyRequest(HttpRequest):
    def __init__(self, *args, **kwargs):
        super(DummyRequest, self).__init__(*args, **kwargs)
        self.method = 'GET'
        self.path = ''
        self.user = User.objects.all()[0]
        self.session = {}
        self._local_site_name = None
        self.local_site = None

        # This is normally set internally by Djblets, but we don't
        # go through the standard __call__ flow.
        self._djblets_webapi_object_cache = {}

    def build_absolute_uri(self, location=None):
        if not self.path and not location:
            return '/api/'

        if not location:
            location = self.path

        if not location.startswith('http://'):
            location = 'http://reviews.example.com' + location

        return location


class ResourceDirective(Directive):
    has_content = True
    required_arguments = 0
    option_spec = {
        'classname': directives.unchanged_required,
        'is-list': directives.flag,
        'hide-links': directives.flag,
        'hide-examples': directives.flag,
    }

    item_http_methods = set(['GET', 'DELETE', 'PUT'])
    list_http_methods = set(['GET', 'POST'])

    FILTERED_MIMETYPES = [
        'application/json',
        'application/xml',
    ]

    def run(self):
        try:
            resource_class = self.get_resource_class(self.options['classname'])
        except ResourceNotFound, e:
            return e.error_node

        # Add the class's file and this extension to the dependencies.
        env = self.state.document.settings.env
        env.note_dependency(__file__)
        env.note_dependency(sys.modules[resource_class.__module__].__file__)

        resource = get_resource_from_class(resource_class)

        is_list = 'is-list' in self.options

        docname = 'webapi2.0-%s-resource' % \
            get_resource_docname(env.app, resource, is_list)
        resource_title = get_resource_title(resource, is_list)

        targetnode = nodes.target('', '', ids=[docname], names=[docname])
        self.state.document.note_explicit_target(targetnode)
        main_section = nodes.section(ids=[docname])

        # Main section
        main_section += nodes.title(text=resource_title)

        for attr_name, text_fmt in (('added_in', 'Added in %s'),
                                    ('deprecated_in', 'Deprecated in %s'),
                                    ('removed_in', 'Removed in %s')):
            version = getattr(resource, attr_name, None)

            if not version:
                if is_list:
                    prefix = 'list_resource'
                else:
                    prefix = 'item_resource'

                version = getattr(resource, '%s_%s' % (prefix, attr_name),
                                  None)

            if version:
                paragraph = nodes.paragraph()
                paragraph += nodes.emphasis(text=text_fmt % version,
                                            classes=['resource-versioning'])

                main_section += paragraph

        main_section += parse_text(
            self, inspect.getdoc(resource),
            where='%s class docstring' % self.options['classname'])

        if getattr(resource, 'required_features', False):
            required_features = nodes.important()
            required_features += nodes.inline(
                text='Using this resource requires extra features to be '
                     'enabled on the server. See "Required Features" below.')
            main_section += required_features

        # Details section
        details_section = nodes.section(ids=['details'])
        main_section += details_section

        details_section += nodes.title(text='Details')
        details_section += self.build_details_table(resource)

        # Fields section
        if (resource.fields and
            (not is_list or resource.singleton)):
            fields_section = nodes.section(ids=['fields'])
            main_section += fields_section

            fields_section += nodes.title(text='Fields')
            fields_section += self.build_fields_table(resource.fields)

        # Links section
        if 'hide-links' not in self.options:
            fields_section = nodes.section(ids=['links'])
            main_section += fields_section

            fields_section += nodes.title(text='Links')
            fields_section += self.build_links_table(resource)

        # HTTP method descriptions
        for http_method in self.get_http_methods(resource, is_list):
            method_section = nodes.section(ids=[http_method])
            main_section += method_section

            method_section += nodes.title(text='HTTP %s' % http_method)
            method_section += self.build_http_method_section(resource,
                                                             http_method)

        if 'hide-examples' not in self.options:
            examples_section = nodes.section(ids=['examples'])
            examples_section += nodes.title(text='Examples')

            has_examples = False

            if is_list:
                mimetype_key = 'list'
            else:
                mimetype_key = 'item'

            for mimetype in resource.allowed_mimetypes:
                try:
                    mimetype = mimetype[mimetype_key]
                except KeyError:
                    continue

                if mimetype in self.FILTERED_MIMETYPES:
                    # Resources have more specific mimetypes. We want to
                    # filter out the general ones (like application/json)
                    # so we don't show redundant examples.
                    continue

                if mimetype.endswith('xml'):
                    # JSON is preferred. While we support XML, let's not
                    # continue to advertise it.
                    continue

                url, headers, data = \
                    self.fetch_resource_data(resource, mimetype)
                example_node = build_example(headers, data, mimetype)

                if example_node:
                    example_section = \
                        nodes.section(ids=['example_' + mimetype],
                                      classes=['examples', 'requests-example'])
                    examples_section += example_section

                    example_section += nodes.title(text=mimetype)

                    accept_mimetype = mimetype

                    if (mimetype.startswith('application/') and
                        mimetype.endswith('+json')):
                        # Instead of telling the user to ask for a specific
                        # mimetype on the request, show them that asking for
                        # application/json works fine.
                        accept_mimetype = 'application/json'

                    curl_text = (
                        '$ curl http://reviews.example.com%s -H "Accept: %s"'
                        % (url, accept_mimetype)
                    )
                    example_section += nodes.literal_block(
                        curl_text, curl_text, classes=['cmdline'])

                    example_section += nodes.literal_block(
                        headers, headers, classes=['http-headers'])
                    example_section += example_node
                    has_examples = True

            if has_examples:
                main_section += examples_section

        return [targetnode, main_section]

    def build_details_table(self, resource):
        env = self.state.document.settings.env
        app = env.app

        is_list = 'is-list' in self.options

        table = nodes.table(classes=['resource-info'])

        tgroup = nodes.tgroup(cols=2)
        table += tgroup

        tgroup += nodes.colspec(colwidth=30, classes=['field'])
        tgroup += nodes.colspec(colwidth=70, classes=['value'])

        tbody = nodes.tbody()
        tgroup += tbody

        # Name
        if is_list:
            resource_name = resource.name_plural
        else:
            resource_name = resource.name

        append_detail_row(tbody, "Name", nodes.literal(text=resource_name))

        # URI
        uri_template = get_resource_uri_template(resource, not is_list)
        append_detail_row(tbody, "URI", nodes.literal(text=uri_template))

        # Required features
        if getattr(resource, 'required_features', False):
            feature_list = nodes.bullet_list()

            for feature in resource.required_features:
                item = nodes.list_item()
                paragraph = nodes.paragraph()

                paragraph += nodes.inline(text=feature.feature_id)
                item += paragraph
                feature_list += item

            append_detail_row(tbody, 'Required Features', feature_list)

        # Token Policy ID
        if hasattr(resource, 'policy_id'):
            append_detail_row(tbody, "Token Policy ID",
                              nodes.literal(text=resource.policy_id))

        # HTTP Methods
        allowed_http_methods = self.get_http_methods(resource, is_list)
        bullet_list = nodes.bullet_list()

        for http_method in allowed_http_methods:
            item = nodes.list_item()
            bullet_list += item

            paragraph = nodes.paragraph()
            item += paragraph

            ref = nodes.reference(text=http_method, refid=http_method)
            paragraph += ref

            doc_summary = self.get_doc_for_http_method(resource, http_method)
            i = doc_summary.find('.')

            if i != -1:
                doc_summary = doc_summary[:i + 1]

            paragraph += nodes.inline(text=" - ")
            paragraph += parse_text(
                self, doc_summary,
                wrapper_node_type=nodes.inline,
                where='HTTP %s handler summary for %s'
                      % (http_method, self.options['classname']))

        append_detail_row(tbody, "HTTP Methods", bullet_list)

        # Parent Resource
        if is_list or resource.uri_object_key is None:
            parent_resource = resource._parent_resource
            is_parent_list = False
        else:
            parent_resource = resource
            is_parent_list = True

        if parent_resource:
            paragraph = nodes.paragraph()
            paragraph += get_ref_to_resource(app, parent_resource,
                                             is_parent_list)
        else:
            paragraph = 'None.'

        append_detail_row(tbody, "Parent Resource", paragraph)

        # Child Resources
        if is_list:
            child_resources = list(resource.list_child_resources)

            if resource.name != resource.name_plural:
                if resource.uri_object_key:
                    child_resources.append(resource)

                are_children_lists = False
            else:
                are_children_lists = True
        else:
            child_resources = resource.item_child_resources
            are_children_lists = True

        if child_resources:
            tocnode = addnodes.toctree()
            tocnode['glob'] = None
            tocnode['maxdepth'] = 1
            tocnode['hidden'] = False

            docnames = sorted([
                docname_join(env.docname,
                             get_resource_docname(app, child_resource,
                                                  are_children_lists))
                for child_resource in child_resources
            ])

            tocnode['includefiles'] = docnames
            tocnode['entries'] = [(None, docname) for docname in docnames]
        else:
            tocnode = nodes.paragraph(text="None")

        append_detail_row(tbody, "Child Resources", tocnode)

        # Anonymous Access
        if is_list and not resource.singleton:
            getter = resource.get_list
        else:
            getter = resource.get

        if getattr(getter, 'login_required', False):
            anonymous_access = 'No'
        elif getattr(getter, 'checks_login_required', False):
            anonymous_access = 'Yes, if anonymous site access is enabled'
        else:
            anonymous_access = 'Yes'

        append_detail_row(tbody, "Anonymous Access", anonymous_access)

        return table

    def build_fields_table(self, fields, required_field_names=None):
        """Build a table representing a list of fields.

        Args:
            fields (dict):
                The fields to display.

            required_field_names (set of unicode, optional):
                The field names that are required.

        Returns:
            list of docutils.nodes.Node:
            The resulting list of nodes for the fields table.
        """
        options = {
            'fields': fields,
        }

        if required_field_names is not None:
            options.update({
                'show-requirement-labels': True,
                'required-field-names': set(required_field_names),
            })

        return run_directive(self, 'webapi-resource-field-list',
                             options=options)

    def build_links_table(self, resource):
        is_list = 'is-list' in self.options

        table = nodes.table()

        tgroup = nodes.tgroup(cols=3)
        table += tgroup

        tgroup += nodes.colspec(colwidth=25)
        tgroup += nodes.colspec(colwidth=15)
        tgroup += nodes.colspec(colwidth=60)

        thead = nodes.thead()
        tgroup += thead
        append_row(thead, ['Name', 'Method', 'Resource'])

        tbody = nodes.tbody()
        tgroup += tbody

        request = DummyRequest()

        if is_list:
            child_resources = resource.list_child_resources
        else:
            child_resources = resource.item_child_resources

        names_to_resource = {}

        for child in child_resources:
            names_to_resource[child.name_plural] = (child, True)

        if not is_list and resource.model:
            child_keys = {}
            create_fake_resource_path(request, resource, child_keys, True)
            obj = resource.get_queryset(request, **child_keys)[0]
        else:
            obj = None

        related_links = resource.get_related_links(request=request, obj=obj)

        for key, info in related_links.iteritems():
            if 'resource' in info:
                names_to_resource[key] = \
                    (info['resource'], info.get('list-resource', False))

        links = resource.get_links(child_resources, request=DummyRequest(),
                                   obj=obj)

        app = self.state.document.settings.env.app

        for linkname in sorted(links.iterkeys()):
            info = links[linkname]
            child, is_child_link = \
                names_to_resource.get(linkname, (resource, is_list))

            paragraph = nodes.paragraph()
            paragraph += get_ref_to_resource(app, child, is_child_link)

            append_row(tbody,
                       [nodes.strong(text=linkname),
                        info['method'],
                        paragraph])

        return table

    def build_http_method_section(self, resource, http_method):
        doc = self.get_doc_for_http_method(resource, http_method)
        http_method_func = self.get_http_method_func(resource, http_method)

        # Description text
        returned_nodes = [
            parse_text(self, doc,
                       wrapper_node_type=nodes.paragraph,
                       where='HTTP %s doc' % http_method),
        ]

        # Request Parameters section
        required_fields = getattr(http_method_func, 'required_fields', [])
        optional_fields = getattr(http_method_func, 'optional_fields', [])

        if required_fields or optional_fields:
            all_fields = dict(required_fields)
            all_fields.update(optional_fields)

            fields_section = nodes.section(ids=['%s_params' % http_method])
            returned_nodes.append(fields_section)

            fields_section += nodes.title(text='Request Parameters')

            table = self.build_fields_table(
                all_fields,
                required_field_names=set(six.iterkeys(required_fields)))
            fields_section += table

        # Errors section
        errors = getattr(http_method_func, 'response_errors', [])

        if errors:
            errors_section = nodes.section(ids=['%s_errors' % http_method])
            returned_nodes.append(errors_section)

            errors_section += nodes.title(text='Errors')
            errors_section += self.build_errors_table(errors)

        return returned_nodes

    def build_errors_table(self, errors):
        """Build a table representing a list of errors.

        Args:
            errors (list of djblets.webapi.errors.WebAPIError):
                The errors to display.

        Returns:
            list of docutils.nodes.Node:
            The resulting list of nodes for the errors table.
        """
        table = nodes.table(classes=['api-errors'])

        tgroup = nodes.tgroup(cols=2)
        table += tgroup

        tgroup += nodes.colspec(colwidth=25)
        tgroup += nodes.colspec(colwidth=75)

        tbody = nodes.tbody()
        tgroup += tbody

        for error in sorted(errors, key=lambda x: x.code):
            http_code = nodes.inline(classes=['http-error'])
            http_code += nodes.reference(
                text='HTTP %s - %s' % (error.http_status,
                                       HTTP_STATUS_CODES[error.http_status]),
                refuri=(DEFAULT_HTTP_STATUS_CODES_URL
                        % error.http_status)),

            error_code = nodes.inline(classes=['api-error'])
            error_code += get_ref_to_error(error)

            error_info = nodes.inline()
            error_info += error_code
            error_info += http_code

            append_row(
                tbody,
                [
                    error_info,
                    nodes.inline(text=error.msg),
                ])

        return table

    def fetch_resource_data(self, resource, mimetype):
        features = {
            feature.feature_id: True
            for feature in resource.required_features
        }

        with override_feature_checks(features):
            kwargs = {}
            request = DummyRequest()
            request.path = create_fake_resource_path(
                request, resource, kwargs, 'is-list' not in self.options)

            headers, data = fetch_response_data(resource, mimetype, request,
                                                **kwargs)

            return request.path, headers, data

    def get_resource_class(self, classname):
        try:
            return get_from_module(classname)
        except ImportError:
            raise ResourceNotFound(self, classname)

    def get_http_method_func(self, resource, http_method):
        if (http_method == 'GET' and 'is-list' in self.options and
            not resource.singleton):
            method_name = 'get_list'
        else:
            method_name = resource.method_mapping[http_method]

            # Change "put" and "post" to "update" and "create", respectively.
            # "put" and "post" are just wrappers and we don't want to show
            # their documentation.
            if method_name == 'put':
                method_name = 'update'
            elif method_name == 'post':
                method_name = 'create'

        return getattr(resource, method_name)

    def get_doc_for_http_method(self, resource, http_method):
        return inspect.getdoc(self.get_http_method_func(resource,
                                                        http_method)) or ''

    def get_http_methods(self, resource, is_list):
        if is_list:
            possible_http_methods = self.list_http_methods
        else:
            possible_http_methods = self.item_http_methods

        return sorted(
            set(resource.allowed_methods).intersection(possible_http_methods))


class ResourceFieldListDirective(Directive):
    """Directive for listing fields in a resource.

    This directive can be used to list the fields belonging to a resource,
    the fields within part of a resource's payload, or fields accepted by
    an operation on a resource.

    The fields can be provided directly (if being called by Python code)
    through the ``fields`` and ``required-field-names`` options. Otherwise,
    this will parse the content of the directive for any
    ``webapi-resource-field`` directives and use those instead.
    """

    has_content = True
    option_spec = {
        'fields': directives.unchanged,
        'required-field-names': directives.unchanged,
    }

    def run(self):
        """Run the directive and render the resulting fields.

        Returns:
            list of docutils.nodes.Node:
            The resulting nodes.
        """
        fields = self.options.get('fields')
        required_fields = self.options.get('required-field-names')

        table = nodes.table(classes=['resource-fields'])

        tgroup = nodes.tgroup(cols=3)
        table += tgroup

        tgroup += nodes.colspec(colwidth=15, classes=['field'])
        tgroup += nodes.colspec(colwidth=85, classes=['description'])

        tbody = nodes.tbody()
        tgroup += tbody

        if fields is not None:
            assert isinstance(fields, dict)

            if required_fields is not None:
                field_keys = sorted(
                    fields.iterkeys(),
                    key=lambda field: (field not in required_fields, field))
            else:
                field_keys = sorted(fields.iterkeys())

            for field in field_keys:
                info = fields[field]

                options = {
                    'name': field,
                    'type': info['type'],
                }

                if info.get('supports_text_types'):
                    options['supports-text-types'] = True

                if required_fields is not None and field in required_fields:
                    options['show-required'] = True

                if info.get('added_in'):
                    options['added-in'] = info['added_in']

                if info.get('deprecated_in'):
                    options['deprecated-in'] = info['deprecated_in']

                if info.get('removed_in'):
                    options['removed-in'] = info['removed_in']

                field_row = run_directive(
                    self,
                    'webapi-resource-field',
                    content='\n'.join(prepare_docstring(info['description'])),
                    options=options)

                tbody += field_row
        elif self.content:
            node = nodes.Element()
            self.state.nested_parse(self.content, self.content_offset,
                                    node)

            # ResourceFieldDirective outputs two fields (two table cells) per
            # field. We want to loop through and grab each.
            tbody += node.children

        return [table]


class ResourceFieldDirective(Directive):
    """Directive for displaying information on a field in a resource.

    This directive can be used to display details about a specific field
    belonging to a resource, a part of a resource's payload, or a field
    accepted by an operation on a resource.

    This is expected to be added into a ``webapi-resource-field-list``
    directive. The resulting node is a table row.
    """

    has_content = True
    option_spec = {
        'name': directives.unchanged_required,
        'type': directives.unchanged_required,
        'show-required': directives.flag,
        'supports-text-types': directives.flag,
        'added-in': directives.unchanged,
        'deprecated-in': directives.unchanged,
        'removed-in': directives.unchanged,
    }

    type_mapping = {
        int: 'Integer',
        str: 'String',
        unicode: 'String',
        bool: 'Boolean',
        dict: 'Dictionary',
        list: 'List',
        file: 'Uploaded File',
    }

    type_name_mapping = {
        'int': int,
        'str': str,
        'unicode': unicode,
        'bool': bool,
        'dict': dict,
        'file': file,
        'list': list,
    }

    def run(self):
        """Run the directive and render the resulting fields.

        Returns:
            list of docutils.nodes.Node:
            The resulting nodes.
        """
        self.assert_has_content()

        name = self.options['name']

        # Field/type information
        field_node = nodes.inline()
        field_node += nodes.strong(text=name, classes=['field-name'])

        type_node = nodes.inline(classes=['field-type'])
        field_node += type_node

        if 'supports-text-types' in self.options:
            type_node += get_ref_to_doc('webapi2.0-text-fields', 'Rich Text')
        else:
            type_node += self._get_type_name(self.options['type'])

        # Description/required/versioning information
        description_node = nodes.inline()

        if 'show-required' in self.options:
            description_node += nodes.inline(text='Required',
                                             classes=['field-required'])

        if 'deprecated-in' in self.options:
            description_node += nodes.inline(text='Deprecated',
                                             classes=['field-deprecated'])

        if isinstance(self.content, StringList):
            description = '\n'.join(self.content)
        else:
            description = self.content

        description_node += parse_text(self, description)

        if 'added-in' in self.options:
            paragraph = nodes.paragraph()
            paragraph += nodes.emphasis(
                text='Added in %s\n' % self.options['added-in'],
                classes=['field-versioning'])
            description_node += paragraph

        if 'deprecated-in' in self.options:
            paragraph = nodes.paragraph()
            paragraph += nodes.emphasis(
                text='Deprecated in %s\n' % self.options['deprecated-in'],
                classes=['field-versioning'])
            description_node += paragraph

        if 'removed-in' in self.options:
            paragraph = nodes.paragraph()
            paragraph += nodes.emphasis(
                text='Removed in %s\n' % self.options['removed-in'],
                classes=['field-versioning'])
            description_node += paragraph

        row = nodes.row()

        entry = nodes.entry()
        entry += field_node
        row += entry

        entry = nodes.entry()
        entry += description_node
        row += entry

        return [row]

    def _get_type_name(self, field_type, nested=False):
        """Return the displayed name for a given type.

        This will attempt to take a type (either a string representation or
        a Python structure) and return a string that can be used for display
        in the API docs.

        This may also be provided a Python class path for a resource.

        Args:
            field_type (object):
                The type of field (as a Python structure), a string
                representing a Python structure, or the class path to a
                resource.

            nested (bool, optional):
                Whether this call is nested within another call to this
                function.

        Returns:
            unicode:
            The resulting string used for display.

        Raises:
            ResourceNotFound:
                A resource path appeared to be provided, but a resource was
                not found.

            ValueError:
                The type is unsupported.
        """
        if isinstance(field_type, basestring) and field_type is not str:
            # First see if this is a string name for a type. This would be
            # coming from a docstring.
            try:
                field_type = self.type_name_mapping[field_type]
            except KeyError:
                if '.' in field_type:
                    # We may be dealing with a forward-declared class.
                    try:
                        field_type = get_from_module(field_type)
                    except ImportError:
                        raise ResourceNotFound(self, field_type)
                else:
                    # Maybe we can parse this?
                    field_type = self._parse_type_string(field_type)

        if type(field_type) is list:
            result = []

            if not nested:
                result.append(nodes.inline(text='List of '))

            if len(field_type) > 1:
                result.append(nodes.inline(text='['))

            first = True

            for item in field_type:
                if not first:
                    result.append(nodes.inline(text=', '))

                result += self._get_type_name(item, nested=True)

                first = False

            if len(field_type) > 1:
                result.append(nodes.inline(text=']'))

            return result
        elif type(field_type) is tuple:
            value_nodes = []

            for value in field_type:
                if value_nodes:
                    value_nodes.append(nodes.inline(text=', '))

                value_nodes.append(nodes.literal(text=value))

            return [nodes.inline(text='One of ')] + value_nodes
        elif (inspect.isclass(field_type) and
              issubclass(field_type, WebAPIResource)):
            return [get_ref_to_resource(self.state.document.settings.env.app,
                                        field_type, False)]
        elif field_type in self.type_mapping:
            return [nodes.inline(text=self.type_mapping[field_type])]
        else:
            raise ValueError('Unsupported type %r' % (field_type,))

    def _parse_type_string(self, type_str):
        """Parse a string representing a given type.

        The string can represent a simple Python primitive (``list``, ``dict``,
        etc.) or a nested structure (``list[dict]``, ``list[[int, unicode]]``,
        etc.).

        Args:
            type_str (unicode):
                The string to parse.

        Returns:
            object
            The resulting Python structure for the given type string.

        Raises:
            ValueError:
                The type is unsupported.
        """
        def _parse_node(node):
            if isinstance(node, ast.Str):
                return node.s
            elif isinstance(node, ast.Num):
                return node.n
            elif isinstance(node, ast.Tuple):
                return tuple(_parse_node(item) for item in node.elts)
            elif isinstance(node, ast.List):
                return list(_parse_node(item) for item in node.elts)
            elif isinstance(node, ast.Dict):
                return dict(
                    (_parse_node(key), _parse_node(value))
                    for key, value in six.iteritems(node.elts)
                )
            elif isinstance(node, ast.Name):
                try:
                    return self.type_name_mapping[node.id]
                except KeyError:
                    raise ValueError(
                        'Unsupported node name "%s" for type string %r'
                        % (node.id, type_str))
            elif isinstance(node, ast.Subscript):
                return _parse_node(node.value)([_parse_node(node.slice.value)])

            raise ValueError('Unsupported node type %r for type string %r'
                             % (node, type_str))

        return _parse_node(ast.parse(type_str, mode='eval').body)


class ResourceTreeDirective(Directive):
    has_content = True

    def run(self):
        bullet_list = nodes.bullet_list()
        self._output_resource(resources.root, bullet_list, True)

        return [bullet_list]

    def _output_resource(self, resource, parent, is_list):
        item = nodes.list_item()
        parent += item

        paragraph = nodes.paragraph()
        item += paragraph

        paragraph += parse_text(
            self,
            ':ref:`%s <%s>`' %
            (get_resource_title(resource, is_list, False),
             'webapi2.0-%s-resource'
             % get_resource_docname(self.state.document.settings.env.app,
                                    resource, is_list)))

        bullet_list = nodes.bullet_list()
        item += bullet_list

        if is_list:
            if resource.uri_object_key:
                self._output_resource(resource, bullet_list, False)

            for child in resource.list_child_resources:
                self._output_resource(child, bullet_list, True)
        else:
            for child in resource.item_child_resources:
                self._output_resource(child, bullet_list, True)


class ErrorDirective(Directive):
    has_content = True
    final_argument_whitespace = True
    option_spec = {
        'instance': directives.unchanged_required,
        'example-data': directives.unchanged,
        'title': directives.unchanged,
    }

    MIMETYPES = [
        'application/json',
    ]

    def run(self):
        try:
            error_obj = self.get_error_object(self.options['instance'])
        except ErrorNotFound, e:
            return e.error_node

        # Add the class's file and this extension to the dependencies.
        self.state.document.settings.env.note_dependency(__file__)
        self.state.document.settings.env.note_dependency(
            sys.modules[error_obj.__module__].__file__)

        docname = 'webapi2.0-error-%s' % error_obj.code
        error_title = self.get_error_title(error_obj)

        targetnode = nodes.target('', '', ids=[docname], names=[docname])
        self.state.document.note_explicit_target(targetnode)
        main_section = nodes.section(ids=[docname])

        # Details section
        main_section += nodes.title(text=error_title)
        main_section += self.build_details_table(error_obj)

        # Example section
        examples_section = nodes.section(ids=['examples'])
        examples_section += nodes.title(text='Examples')
        extra_params = {}

        if 'example-data' in self.options:
            extra_params = json.loads(self.options['example-data'])

        has_examples = False

        for mimetype in self.MIMETYPES:
            headers, data = \
                fetch_response_data(WebAPIResponseError, mimetype,
                                    err=error_obj,
                                    extra_params=extra_params)
            example_node = build_example(headers, data, mimetype)

            if example_node:
                example_section = nodes.section(ids=['example_' + mimetype])
                examples_section += example_section

                example_section += nodes.title(text=mimetype)
                example_section += example_node
                has_examples = True

        if has_examples:
            main_section += examples_section

        return [targetnode, main_section]

    def build_details_table(self, error_obj):
        table = nodes.table()

        tgroup = nodes.tgroup(cols=2)
        table += tgroup

        tgroup += nodes.colspec(colwidth=20)
        tgroup += nodes.colspec(colwidth=80)

        tbody = nodes.tbody()
        tgroup += tbody

        # API Error Code
        append_detail_row(tbody, 'API Error Code',
                          nodes.literal(text=error_obj.code))

        # HTTP Status Code
        ref = parse_text(self, ':http:`%s`' % error_obj.http_status)
        append_detail_row(tbody, 'HTTP Status Code', ref)

        # Error Text
        append_detail_row(tbody, 'Error Text',
                          nodes.literal(text=error_obj.msg))

        if error_obj.headers:
            if callable(error_obj.headers):
                headers = error_obj.headers(DummyRequest())

            # HTTP Headers
            if len(headers) == 1:
                content = nodes.literal(text=headers.keys()[0])
            else:
                content = nodes.bullet_list()

                for header in headers.iterkeys():
                    item = nodes.list_item()
                    content += item

                    literal = nodes.literal(text=header)
                    item += literal

            append_detail_row(tbody, 'HTTP Headers', content)

        # Description
        append_detail_row(
            tbody, 'Description',
            parse_text(self, '\n'.join(self.content),
                       where='API error %s description' % error_obj.code))

        return table

    def get_error_title(self, error_obj):
        if 'title' in self.options:
            error_title = self.options['title']
        else:
            name = self.options['instance'].split('.')[-1]
            error_title = name.replace('_', ' ').title()

        return '%s - %s' % (error_obj.code, error_title)

    def get_error_object(self, name):
        try:
            return get_from_module(name)
        except ImportError:
            raise ErrorNotFound(self, name)


def parse_text(directive, text, wrapper_node_type=None, where=None):
    """Parse text in ReST format and return a node with the content.

    Args:
        directive (docutils.parsers.rst.Directive):
            The directive that will contain the resulting nodes.

        text (unicode):
            The text to parse.

        wrapper_node_type (docutils.nodes.Node, optional):
            An optional node type used to contain the children.

        where (unicode, optional):
            Information on the location being parsed in case there's a
            failure.

    Returns:
        list of docutils.nodes.Node:
        The resulting list of parsed nodes.
    """
    assert text is not None, 'Missing text during parse_text in %s' % where

    if wrapper_node_type:
        node_type = wrapper_node_type
    else:
        node_type = nodes.container

    node = node_type(rawsource=text)
    directive.state.nested_parse(ViewList(string2lines(text), source=''),
                                 0, node)

    if wrapper_node_type:
        return node
    else:
        return node.children


def run_directive(parent_directive, name, content='', options={}):
    """Run and render a directive.

    Args:
        parent_directive (docutils.parsers.rst.Directive):
            The directive running another directive.

        name (unicode):
            The name of the directive to run.

        content (unicode, optional):
            The content to pass to the directive.

        options (dict, optional):
            The options to pass to the directive.

    Returns:
        list of docutils.nodes.Node:
        The resulting list of nodes from the directive.
    """
    state = parent_directive.state
    directive_class, messages = directives.directive(name,
                                                     state.memo.language,
                                                     state.document)
    state.parent += messages

    if not directive_class:
        return state.unknown_directive(name)

    state_machine = state.state_machine
    lineno = state_machine.abs_line_number()

    directive = directive_class(
        name=name,
        arguments=[],
        options=options,
        content=content,
        lineno=lineno,
        content_offset=0,
        block_text='',
        state=parent_directive.state,
        state_machine=state_machine)

    try:
        return directive.run()
    except DirectiveError as e:
        return [
            parent_directive.reporter.system_message(e.level, e.msg,
                                                     line=lineno),
        ]


def get_from_module(name):
    i = name.rfind('.')
    module, attr = name[:i], name[i + 1:]

    try:
        mod = __import__(module.encode('utf-8'), {}, {},
                         [attr.encode('utf-8')])
        return getattr(mod, attr)
    except AttributeError:
        raise ImportError('Unable to load "%s" from "%s"' % (attr, module))


def append_row(tbody, cells):
    row = nodes.row()
    tbody += row

    for cell in cells:
        entry = nodes.entry()
        row += entry

        if isinstance(cell, basestring):
            node = nodes.paragraph(text=cell)
        else:
            node = cell

        entry += node


def append_detail_row(tbody, header_text, detail):
    header_node = nodes.strong(text=header_text)

    if isinstance(detail, basestring):
        detail_node = [nodes.paragraph(text=text)
                       for text in detail.split('\n\n')]
    else:
        detail_node = detail

    append_row(tbody, [header_node, detail_node])


FIRST_CAP_RE = re.compile(r'(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile(r'([a-z0-9])([A-Z])')


def uncamelcase(name, separator='_'):
    """
    Converts a string from CamelCase into a lowercase name separated by
    a provided separator.
    """
    s1 = FIRST_CAP_RE.sub(r'\1%s\2' % separator, name)
    return ALL_CAP_RE.sub(r'\1%s\2' % separator, s1).lower()


def get_resource_title(resource, is_list, append_resource=True):
    """Returns a human-readable name for the resource."""
    if hasattr(resource, 'verbose_name'):
        normalized_title = resource.verbose_name
    else:
        class_name = resource.__class__.__name__
        class_name = class_name.replace('Resource', '')
        normalized_title = title(uncamelcase(class_name, ' '))

    if is_list:
        s = '%s List' % normalized_title
    else:
        s = normalized_title

    if append_resource:
        s += ' Resource'

    return s


def get_resource_docname(app, resource, is_list):
    """Returns the name of the page used for a resource's documentation."""
    if inspect.isclass(resource):
        class_name = resource.__name__
    else:
        class_name = resource.__class__.__name__

    class_name = class_name.replace('Resource', '')
    docname = uncamelcase(class_name, '-')
    docname = app.config.webapi_docname_map.get(docname, docname)

    if is_list and resource.name != resource.name_plural:
        docname = '%s-list' % docname

    return docname


def get_ref_to_doc(refname, title=''):
    """Returns a node that links to a document with the given ref name."""
    ref = addnodes.pending_xref(reftype='ref', reftarget=refname,
                                refexplicit=(title != ''), refdomain='std')
    ref += nodes.literal(title, title, classes=['xref'])
    return ref


def get_ref_to_resource(app, resource, is_list):
    """Returns a node that links to a resource's documentation."""
    return get_ref_to_doc('webapi2.0-%s-resource' %
                          get_resource_docname(app, resource, is_list))


def get_ref_to_error(error, title=''):
    """Returns a node that links to an error's documentation."""
    return get_ref_to_doc('webapi2.0-error-%s' % error.code,
                          title=title)


def get_resource_uri_template(resource, include_child):
    """Returns the URI template for a resource.

    This will go up the resource tree, building a URI based on the URIs
    of the parents.
    """
    if resource.name == 'root':
        path = '/api/'
    else:
        if resource._parent_resource:
            path = get_resource_uri_template(resource._parent_resource, True)

        path += '%s/' % resource.uri_name

        if not resource.singleton and include_child and resource.model:
            path += '{%s}/' % resource.uri_object_key

    return path


def create_fake_resource_path(request, resource, child_keys, include_child):
    """Creates a fake path to a resource.

    This will go up the resource tree, building a URI based on the URIs
    of the parents and based on objects sitting in the database.
    """
    if resource._parent_resource and resource._parent_resource.name != "root":
        path = create_fake_resource_path(request, resource._parent_resource,
                                         child_keys, True)
    else:
        path = '/api/'

    if resource.name != 'root':
        path += '%s/' % resource.uri_name

        if (not resource.singleton and
            include_child and
            resource.model and
            resource.uri_object_key):
                q = resource.get_queryset(request, **child_keys)

                if q.count() == 0:
                    logging.critical('Resource "%s" requires objects in the '
                                     'database. Tried with URL child keys: %r',
                                     resource.__class__, child_keys)

                    # Do the assert so it shows up in the logs.
                    assert q.count() > 0

                obj = q[0]
                value = getattr(obj, resource.model_object_key)
                child_keys[resource.uri_object_key] = value
                path += '%s/' % value

    return path


def build_example(headers, data, mimetype):
    if not data:
        return None

    language = None

    for base_mimetype, lang in MIMETYPE_LANGUAGES:
        if is_mimetype_a(mimetype, base_mimetype):
            language = lang
            break

    if language == 'javascript':
        code = json.dumps(json.loads(data), sort_keys=True, indent=2)
    else:
        code = data

    return nodes.literal_block(code, code, language=language or 'text',
                               classes=['example-payload'])


def fetch_response_data(response_class, mimetype, request=None, **kwargs):
    if not request:
        request = DummyRequest()

    request.META['HTTP_ACCEPT'] = mimetype

    result = unicode(response_class(request, **kwargs))
    return result.split('\r\n\r\n', 1)


def setup(app):
    app.add_config_value(b'webapi_docname_map', {}, b'env')

    app.add_directive('webapi-resource', ResourceDirective)
    app.add_directive('webapi-resource-field-list', ResourceFieldListDirective)
    app.add_directive('webapi-resource-field', ResourceFieldDirective)
    app.add_directive('webapi-resource-tree', ResourceTreeDirective)
    app.add_directive('webapi-error', ErrorDirective)
    app.add_crossref_type('webapi2.0', 'webapi2.0', 'single: %s',
                          nodes.emphasis)

    # Filter out some additional log messages.
    for name in ('djblets.util.templatetags.djblets_images',):
        logging.getLogger(name).disabled = True



"""A backend for the Elasticsearch search engine."""

from __future__ import unicode_literals

from importlib import import_module

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext, ugettext_lazy as _

from reviewboard.search.search_backends.base import (SearchBackend,
                                                     SearchBackendForm)


class ElasticsearchConfigForm(SearchBackendForm):
    """A form for configuring the Elasticsearch search backend."""

    url = forms.URLField(
        label=_('Elasticsearch URL'),
        help_text=_('The URL of the Elasticsearch server.'),
        widget=forms.TextInput(attrs={'size': 80}))

    index_name = forms.CharField(
        label=_('Elasticsearch index name'),
        help_text=_('The name of the Elasticsearch index.'),
        widget=forms.TextInput(attrs={'size': 40}))


class ElasticsearchBackend(SearchBackend):
    """A search backend for integrating with Elasticsearch"""

    search_backend_id = 'elasticsearch'
    name = _('Elasticsearch')
    haystack_backend_name = ('haystack.backends.elasticsearch_backend.'
                             'ElasticsearchSearchEngine')
    default_settings = {
        'URL': 'http://127.0.0.1:9200/',
        'INDEX_NAME': 'reviewboard',
    }
    config_form_class = ElasticsearchConfigForm
    form_field_map = {
        'url': 'URL',
        'index_name': 'INDEX_NAME',
    }

    def validate(self):
        """Ensure that the elasticsearch Python module is installed.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the ``elasticsearch`` module is not installed.
        """
        try:
            module = import_module('elasticsearch')
        except ImportError:
            module = None

        # Check whether there's a supported version of the module available.
        # Note that technically, elasticsearch 1.x is supported, but it's
        # pretty old. If we're going to reference a version, we want to
        # reference 2.x.
        if (module is None or
            not hasattr(module, 'VERSION') or
            module.VERSION[0] > 2):
            raise ValidationError(ugettext(
                'The elasticsearch 2.x Python module (and the Elasticsearch '
                '2.x service) is required. The module can be installed by '
                'running: pip install "elasticsearch>=2.0,<3.0"'
            ))



"""Unit tests for reviewboard.reviews.views.PreviewReviewRequestEmailView."""

from __future__ import unicode_literals

from django.contrib.auth.models import User

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class PreviewReviewRequestEmailViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.PreviewReviewRequestEmailView.
    """

    fixtures = ['test_users', 'test_scmtools']

    def test_access_with_debug(self):
        """Testing PreviewReviewRequestEmailView access with DEBUG=True"""
        review_request = self.create_review_request(publish=True)

        with self.settings(DEBUG=True):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-request-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 200)

    def test_access_without_debug(self):
        """Testing PreviewReviewRequestEmailView access with DEBUG=False"""
        review_request = self.create_review_request(publish=True)

        with self.settings(DEBUG=False):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-request-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 404)

    def test_with_valid_change_id(self):
        """Testing PreviewReviewRequestEmailView access with valid change ID"""
        user = User.objects.create_user('test_user')
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    target_people=[user])

        self.create_diffset(review_request, draft=True)
        review_request.publish(review_request.submitter)

        with self.settings(DEBUG=True):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-request-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'message_format': 'text',
                        'changedesc_id': review_request.changedescs.get().pk,
                    }))

        self.assertEqual(response.status_code, 200)

    def test_with_invalid_change_id(self):
        """Testing PreviewReviewRequestEmailView access with invalid change ID
        """
        user = User.objects.create_user('test_user')
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    target_people=[user])

        self.create_diffset(review_request, draft=True)
        review_request.publish(review_request.submitter)

        with self.settings(DEBUG=True):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-request-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'message_format': 'text',
                        'changedesc_id': 100,
                    }))

        self.assertEqual(response.status_code, 404)



"""Unit tests for the ReviewBoardGateway hosting service."""

from __future__ import unicode_literals

import hashlib
import hmac

from djblets.testing.decorators import add_fixtures

from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import encrypt_password
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse


class ReviewBoardGatewayTestCase(HostingServiceTestCase):
    """Base test case for the ReviewBoardGateway hosting service."""

    service_name = 'rbgateway'

    default_use_hosting_url = True
    default_account_data = {
        'private_token': encrypt_password('abc123'),
    }

    default_repository_extra_data = {
        'rbgateway_repo_name': 'myrepo',
    }


class ReviewBoardGatewayTests(ReviewBoardGatewayTestCase):
    """Unit tests for the ReviewBoardGateway hosting service."""

    def test_service_support(self):
        """Testing ReviewBoardGateway service support capabilities"""
        self.assertTrue(self.service_class.supports_repositories)
        self.assertTrue(self.service_class.supports_post_commit)
        self.assertFalse(self.service_class.supports_bug_trackers)
        self.assertFalse(self.service_class.supports_ssh_key_association)

    def test_repo_field_values(self):
        """Testing ReviewBoardGateway.get_repository_fields for Git"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'hosting_url': 'https://example.com',
                    'rbgateway_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'https://example.com/repos/myrepo/path',
            })

    def test_authorize(self):
        """Testing ReviewBoardGateway.authorize"""
        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        with self.setup_http_test(payload=b'{"private_token": "abc123"}',
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(username='myuser',
                                  password='mypass',
                                  hosting_url='https://example.com')

        self.assertTrue(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://example.com/session',
            method='POST',
            body='',
            headers={
                'Content-Length': '0',
            })

    def test_check_repository(self):
        """Testing ReviewBoardGateway.check_repository"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(
                path='https://example.com/repos/myrepo/path')

        ctx.assertHTTPCall(
            0,
            url='https://example.com/repos/myrepo/path',
            username=None,
            password=None,
            headers={
                'PRIVATE-TOKEN': 'abc123',
            })

    def test_get_branches_git(self):
        """Testing ReviewBoardGateway.get_branches for a Git repository"""
        payload = self.dump_json([
            {
                'name': 'master',
                'id': 'c272edcac05b00e15440d6274723b639e3acbd7c',
            },
            {
                'name': 'im_a_branch',
                'id': '83904e6acb60e7ec0dcaae6c09a579ab44d0cf38',
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://example.com/repos/myrepo/branches',
            username=None,
            password=None,
            headers={
                'PRIVATE-TOKEN': 'abc123',
            })

        self.assertEqual(
            branches,
            [
                Branch(id='master',
                       commit='c272edcac05b00e15440d6274723b639e3acbd7c',
                       default=True),
                Branch(id='im_a_branch',
                       commit='83904e6acb60e7ec0dcaae6c09a579ab44d0cf38'),
            ])

    def test_get_branches_hg(self):
        """Testing ReviewBoardGateway.get_branches for an Hg repository"""
        payload = self.dump_json([
            {
                'name': 'default',
                'id': '9b1153b8a8eb2f7b1661ed7695c432f5a2b25729',
            },
            {
                'name': 'some-bookmark',
                'id': '0731875ed7a14bdd53503b27b30a08a0452068cf',
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name='Mercurial')
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url='https://example.com/repos/myrepo/branches',
            username=None,
            password=None,
            headers={
                'PRIVATE-TOKEN': 'abc123',
            })

        self.assertEqual(
            branches,
            [
                Branch(id='default',
                       commit='9b1153b8a8eb2f7b1661ed7695c432f5a2b25729',
                       default=True),
                Branch(id='some-bookmark',
                       commit='0731875ed7a14bdd53503b27b30a08a0452068cf'),
            ])


    def test_get_commits(self):
        """Testing ReviewBoardGateway.get_commits"""
        payload = self.dump_json([
            {
                'author': 'Author 1',
                'id': 'bfdde95432b3af879af969bd2377dc3e55ee46e6',
                'date': '2015-02-13 22:34:01 -0700',
                'message': 'Message 1',
                'parent_id': '304c53c163aedfd0c0e0933776f09c24b87f5944',
            },
            {
                'author': 'Author 2',
                'id': '304c53c163aedfd0c0e0933776f09c24b87f5944',
                'date': '2015-02-13 22:32:42 -0700',
                'message': 'Message 2',
                'parent_id': 'fa1330719893098ae397356e8125c2aa45b49221',
            },
            {
                'author': 'Author 3',
                'id': 'fa1330719893098ae397356e8125c2aa45b49221',
                'date': '2015-02-12 16:01:48 -0700',
                'message': 'Message 3',
                'parent_id': '',
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            commits = ctx.service.get_commits(
                repository=repository,
                branch='bfdde95432b3af879af969bd2377dc3e55ee46e6')

        ctx.assertHTTPCall(
            0,
            url=('https://example.com/repos/myrepo/branches/'
                 'bfdde95432b3af879af969bd2377dc3e55ee46e6/commits'),
            username=None,
            password=None,
            headers={
                'PRIVATE-TOKEN': 'abc123',
            })

        self.assertEqual(
            commits,
            [
                Commit(author_name='Author 1',
                       date='2015-02-13 22:34:01 -0700',
                       id='bfdde95432b3af879af969bd2377dc3e55ee46e6',
                       message='Message 1',
                       parent='304c53c163aedfd0c0e0933776f09c24b87f5944'),
                Commit(author_name='Author 2',
                       date='2015-02-13 22:32:42 -0700',
                       id='304c53c163aedfd0c0e0933776f09c24b87f5944',
                       message='Message 2',
                       parent='fa1330719893098ae397356e8125c2aa45b49221'),
                Commit(author_name='Author 3',
                       date='2015-02-12 16:01:48 -0700',
                       id='fa1330719893098ae397356e8125c2aa45b49221',
                       message='Message 3',
                       parent=''),
            ])

        for commit in commits:
            self.assertIsNone(commit.diff)

    def test_get_change(self):
        """Testing ReviewBoardGateway.get_change"""
        diff = (
            'diff --git a/test b/test\n'
            'index 9daeafb9864cf43055ae93beb0afd6c7d144bfa4..'
            'dced80a85fe1e8f13dd5ea19923e5d2e8680020d 100644\n'
            '--- a/test\n'
            '+++ b/test\n'
            '@@ -1 +1,3 @@\n'
            ' test\n'
            '+\n'
            '+test\n'
        )

        payload = self.dump_json({
            'author': 'Some Author',
            'id': 'bfdde95432b3af879af969bd2377dc3e55ee46e6',
            'date': '2015-02-13 22:34:01 -0700',
            'message': 'My Message',
            'parent_id': '304c53c163aedfd0c0e0933776f09c24b87f5944',
            'diff': diff,
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            change = ctx.service.get_change(
                repository=repository,
                revision='bfdde95432b3af879af969bd2377dc3e55ee46e6')

        ctx.assertHTTPCall(
            0,
            url=('https://example.com/repos/myrepo/commits/'
                 'bfdde95432b3af879af969bd2377dc3e55ee46e6'),
            username=None,
            password=None,
            headers={
                'PRIVATE-TOKEN': 'abc123',
            })

        self.assertEqual(
            change,
            Commit(author_name='Some Author',
                   date='2015-02-13 22:34:01 -0700',
                   id='bfdde95432b3af879af969bd2377dc3e55ee46e6',
                   message='My Message',
                   parent='304c53c163aedfd0c0e0933776f09c24b87f5944'))
        self.assertEqual(change.diff, diff.encode('utf-8'))


class CloseSubmittedHookTests(ReviewBoardGatewayTestCase):
    """Unit tests for ReviewBoardGateway's close-submitted hook."""

    fixtures = ['test_users', 'test_scmtools']

    def test_close_submitted_hook_git(self):
        """Testing the ReviewBoardGateway close-submitted hook with a Git
        repository
        """
        self._test_post_commit_hook(tool_name='Git')

    def test_close_submiteed_hook_git_unpublished(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        unpublished review request in a Git repository
        """
        self._test_post_commit_hook(tool_name='Git', publish=False)

    @add_fixtures(['test_site'])
    def test_close_submiteed_hook_git_local_site_unpublished(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        unpublished review request in a Git repository on a Local Site
        """
        self._test_post_commit_hook(
            tool_name='Git',
            local_site=LocalSite.objects.get(name=self.local_site_name),
            publish=False)

    def test_close_submitted_hook_git_tag_target(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Git repository and a tag target
        """
        self._test_post_commit_hook(
            tool_name='Git',
            expected_close_msg='Pushed to release-1.0.7 (bbbbbbb)',
            target_tags=['release-1.0.7', 'some-tag'])

    def test_close_submitted_hook_git_no_target(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Git repository and no target information
        """
        self._test_post_commit_hook(
            tool_name='Git',
            expected_close_msg='Pushed to bbbbbbb',
            target_branch=None)

    def test_close_submitted_hook_hg(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Mercurial repository
        """
        self._test_post_commit_hook(tool_name='Mercurial')

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_hg_local_site(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Mercurial repository on a Local Site
        """
        self._test_post_commit_hook(
            tool_name='Mercurial',
            local_site=LocalSite.objects.get(name=self.local_site_name))

    def test_close_submitted_hook_hg_unpublished(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        unpublished review request in a Mercurial repository
        """
        self._test_post_commit_hook(tool_name='Mercurial', publish=False)

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_hg_local_site_unpublished(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        unpublished review request in a Mercurial repository on a Local Site
        """
        self._test_post_commit_hook(
            tool_name='Mercurial',
            local_site=LocalSite.objects.get(name=self.local_site_name),
            publish=False)

    def test_close_submitted_hook_hg_bookmark_target(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Mercurial repository and a bookmark target
        """
        self._test_post_commit_hook(
            tool_name='Mercurial',
            expected_close_msg='Pushed to dev-work (bbbbbbb)',
            target_bookmarks=['dev-work'])

    def test_close_submitted_hook_hg_tag_target(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Mercurial repository and a tag target
        """
        self._test_post_commit_hook(
            tool_name='Mercurial',
            expected_close_msg='Pushed to @ (bbbbbbb)',
            target_branch='default',
            target_tags=['@', 'tip'])

    def test_close_submitted_hook_hg_no_target(self):
        """Testing the ReviewBoardGateway close-submitted hook with an
        Mercurial repository and no target information
        """
        self._test_post_commit_hook(
            tool_name='Mercurial',
            expected_close_msg='Pushed to bbbbbbb',
            target_branch=None)

    def test_close_submitted_hook_invalid_signature(self):
        """Testing the ReviewBoardGateway close-submitted hook with an invalid
        signature
        """
        account = self.create_hosting_account()
        repository = self.create_repository(tool_name='Git',
                                            hosting_account=account)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=None,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        payload = self.dump_json({
            'event': 'push',
            'commits': [],
        })
        signature = hmac.new(
            b'this is not the secret key',
            payload,
            hashlib.sha1).hexdigest()

        rsp = self.client.post(
            url,
            payload,
            content_type='application/x-www-form-urlencoded',
            HTTP_X_RBG_SIGNATURE=signature,
            HTTP_X_RBG_EVENT='push')

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(rsp.content, 'Bad signature.')

    def test_close_submitted_hook_malformed_payload(self):
        """Testing the ReviewBoardGateway close-submitted hook with a malformed
        signature
        """
        account = self.create_hosting_account()
        repository = self.create_repository(tool_name='Git',
                                            hosting_account=account)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=None,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        payload = 'event=push&commit_id=bbbbbbb&branch=master'
        signature = hmac.new(
            bytes(repository.get_or_create_hooks_uuid()),
            payload,
            hashlib.sha1).hexdigest()

        rsp = self.client.post(
            url,
            payload,
            content_type='application/x-www-form-urlencoded',
            HTTP_X_RBG_SIGNATURE=signature,
            HTTP_X_RBG_EVENT='push')

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(rsp.content, 'Invalid payload format.')

    def test_close_submitted_hook_incomplete_payload(self):
        account = self.create_hosting_account()
        repository = self.create_repository(tool_name='Git',
                                            hosting_account=account)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=None,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        payload = self.dump_json({
            'event': 'push',
        })
        signature = hmac.new(
            bytes(repository.get_or_create_hooks_uuid()),
            payload,
            hashlib.sha1).hexdigest()

        rsp = self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_RBG_SIGNATURE=signature,
            HTTP_X_RBG_EVENT='push')

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(rsp.content, 'Invalid payload; expected "commits".')

    def test_close_submitted_hook_invalid_event(self):
        """Testing the ReviewBoardGateway close-submitted hook endpoint with an
        invalid event
        """
        account = self.create_hosting_account()
        repository = self.create_repository(tool_name='Git',
                                            hosting_account=account)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=None,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        payload = self.dump_json({
            'event': 'unknown-event',
            'repository': 'foo',
        })
        signature = hmac.new(
            bytes(repository.get_or_create_hooks_uuid()),
            payload,
            hashlib.sha1).hexdigest()

        rsp = self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_RBG_SIGNATURE=signature,
            HTTP_X_RBG_EVENT='unknown-event')

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(rsp.content,
                         'Only "ping" and "push" events are supported.')

    def test_ping_event(self):
        """Testing the ReviewBoardGateway close submitted hook endpoint with
        event=ping
        """
        account = self.create_hosting_account()
        repository = self.create_repository(tool_name='Git',
                                            hosting_account=account)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=None,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        payload = self.dump_json({
            'event': 'ping',
            'repository': 'foo',
        })
        signature = hmac.new(
            bytes(repository.get_or_create_hooks_uuid()),
            payload,
            hashlib.sha1).hexdigest()

        rsp = self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_RBG_SIGNATURE=signature,
            HTTP_X_RBG_EVENT='ping')

        self.assertEqual(rsp.status_code, 200)
        self.assertEqual(rsp.content, '')

    def _test_post_commit_hook(self, tool_name, local_site=None, publish=True,
                               expected_close_msg='Pushed to master (bbbbbbb)',
                               **kwargs):
        """Testing posting to a commit hook.

        This will simulate pushing a commit and posting the resulting webhook
        payload from RB Gateway to the handler for the hook.

        Args:
            tool_name (unicode):
                The name of the SCM tool to use.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning the review request.

            publish (bool):
                Whether or not to use a published review request.
        """
        account = self.create_hosting_account(local_site=local_site)
        repository = self.create_repository(tool_name=tool_name,
                                            hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    local_site=local_site,
                                                    publish=publish)

        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'rbgateway-hooks-close-submitted',
            local_site=local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid(),
            **kwargs)
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.SUBMITTED)
        self.assertEqual(review_request.changedescs.count(), 1)

        changedesc = review_request.changedescs.get()
        self.assertEqual(changedesc.text, expected_close_msg)

    def _post_commit_hook_payload(self, url, review_request, secret,
                                  event='push', target_branch='master',
                                  target_bookmarks=None, target_tags=None):
        """Post a payload for a hook for testing.

        Args:
            url (unicode):
                The URL to post to.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being represented in the payload.

            secret (unicode):
                The HMAC secret for the message.

            event (unicode, optional):
                The webhook event.

        Results:
            django.core.handlers.wsgi.WSGIRequest:
            The post request.
        """
        target = {}

        if target_branch is not None:
            target['branch'] = target_branch

        if target_bookmarks is not None:
            target['bookmarks'] = target_bookmarks

        if target_tags is not None:
            target['tags'] = target_tags

        payload = self.dump_json({
            'event': event,
            'repository': review_request.repository.name,
            'commits': [
                {
                    'id': 'b' * 40,
                    'message': (
                        'Commit message.\n\n'
                        'Reviewed at http://example.com%s'
                    ) % review_request.get_absolute_url(),
                    'target': target,
                },
            ],
        })

        signature = hmac.new(bytes(secret), payload, hashlib.sha1).hexdigest()

        return self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_RBG_EVENT=event,
            HTTP_X_RBG_SIGNATURE=signature)



from __future__ import unicode_literals

import os

from django.contrib.auth.models import AnonymousUser, User

from reviewboard.reviews.models import Group
from reviewboard.scmtools.forms import RepositoryForm
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class PolicyTests(TestCase):
    """Unit tests for access policies."""

    fixtures = ['test_scmtools']

    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password='',
                                             email='user@example.com')
        self.anonymous = AnonymousUser()
        self.repo = Repository.objects.create(
            name='test',
            path='example.com:/cvsroot/test',
            username='anonymous',
            tool=Tool.objects.get(name='CVS'))

    def test_repository_public(self):
        """Testing access to a public repository"""
        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertTrue(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertTrue(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_denied(self):
        """Testing no access to an inaccessible private repository"""
        self.repo.public = False
        self.repo.save()

        self.assertFalse(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertNotIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_allowed_by_user(self):
        """Testing access to a private repository accessible by user"""
        self.repo.users.add(self.user)
        self.repo.public = False
        self.repo.save()

        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_allowed_by_review_group(self):
        """Testing access to a private repository accessible by review group"""
        group = Group.objects.create(name='test-group')
        group.users.add(self.user)

        self.repo.public = False
        self.repo.review_groups.add(group)
        self.repo.save()

        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_form_with_local_site_and_bad_group(self):
        """Testing adding a Group to a RepositoryForm with the wrong LocalSite
        """
        test_site = LocalSite.objects.create(name='test')
        tool = Tool.objects.get(name='Subversion')
        group = Group.objects.create(name='test-group')

        svn_repo_path = 'file://' + os.path.join(os.path.dirname(__file__),
                                                 '..', 'testdata', 'svn_repo')

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'review_groups': [group.pk],
            'local_site': test_site.pk,
            'tool': tool.pk,
        })
        self.assertFalse(form.is_valid())

        group.local_site = test_site
        group.save()

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'review_groups': [group.pk],
            'tool': tool.pk,
        })
        self.assertFalse(form.is_valid())

    def test_repository_form_with_local_site_and_bad_user(self):
        """Testing adding a User to a RepositoryForm with the wrong LocalSite
        """
        test_site = LocalSite.objects.create(name='test')
        tool = Tool.objects.get(name='Subversion')

        svn_repo_path = 'file://' + os.path.join(os.path.dirname(__file__),
                                                 '..', 'testdata', 'svn_repo')

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'users': [self.user.pk],
            'local_site': test_site.pk,
            'tool': tool.pk,
        })
        self.assertFalse(form.is_valid())



"""Unit tests for reviewboard.accounts.forms.pages.AccountSettingsForm."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import RequestFactory
from django.views.generic.base import View
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.forms.pages import AccountSettingsForm
from reviewboard.accounts.models import Profile
from reviewboard.accounts.pages import AccountSettingsPage
from reviewboard.testing import TestCase


class AccountSettingsFormTests(TestCase):
    """Unit tests for reviewboard.accounts.forms.pages.AccountSettingsForm."""

    @classmethod
    def setUpClass(cls):
        """Set up the test class."""
        super(AccountSettingsFormTests, cls).setUpClass()

        cls._request_factory = RequestFactory()
        cls._middleware = [
            SessionMiddleware(),
            MessageMiddleware(),
        ]

    @add_fixtures(['test_users'])
    def test_save_syntax_highlighting_disabled(self):
        """Testing AccountSettingsForm.save() with
        diffviewer_syntax_highlighting disabled
        """
        view = View()
        user = User.objects.get(username='doc')

        profile = user.get_profile()
        profile.syntax_highlighting = True
        profile.save()

        request = self._build_request(user)
        page = AccountSettingsPage(view, request, user)

        settings = {'diffviewer_syntax_highlighting': False}

        with self.siteconfig_settings(settings):
            form = AccountSettingsForm(page, request, user, data={
                'syntax_highlighting': False,
                'timezone': profile.timezone,
            })

            self.assertTrue(form.is_valid())
            form.save()

        profile = Profile.objects.get(pk=profile.pk)
        self.assertTrue(profile.syntax_highlighting)

    def _build_request(self, user=None):
        """Make a request and process it through middleware.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user for the request to be authenticated for.

                If not provided, an
                :py:class:`~django.contrib.auth.models.AnonymousUser`
                will be assigned instead.

        Returns:
            django.http.HttpRequest:
            The created request.
        """
        if user is None:
            user = AnonymousUser()

        request = self._request_factory.request()
        request.user = user

        for middleware in self._middleware:
            middleware.process_request(request)

        return request



"""Unit tests for reviewboard.accounts.forms.pages.ProfileForm."""

from __future__ import unicode_literals

from django.contrib import messages
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from kgb import SpyAgency

from reviewboard.accounts.backends import (AuthBackend,
                                           get_enabled_auth_backends)
from reviewboard.accounts.forms.pages import ProfileForm
from reviewboard.accounts.models import Profile
from reviewboard.testing import TestCase


class SandboxAuthBackend(AuthBackend):
    """Mock authentication backend to test extension sandboxing."""

    backend_id = 'test-id'
    name = 'test'
    supports_change_name = True
    supports_change_email = True

    def update_name(self, user):
        """Raise an exception to test sandboxing."""
        raise Exception

    def update_email(self, user):
        """Raise an exception to test sandboxing."""
        raise Exception


class ProfileFormTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.accounts.forms.pages.ProfileForm."""

    def setUp(self):
        super(ProfileFormTests, self).setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('test')
        self.user = User.objects.create_user(username='reviewboard', email='',
                                             password='password')
        self.profile = Profile.objects.get_or_create(user=self.user)
        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [SandboxAuthBackend()])

        # Suppresses MessageFailure Exception at the end of save()
        self.spy_on(messages.add_message,
                    call_fake=lambda *args, **kwargs: None)

    def test_update_name_auth_backend(self):
        """Testing ProfileForm.save with error in auth_backend.update_name"""
        form = ProfileForm(page=None,
                           request=self.request,
                           user=self.user)
        form.cleaned_data = {
            'first_name': 'Barry',
            'last_name': 'Allen',
            'email': 'flash@example.com',
            'profile_private': '',
        }
        self.user.email = 'flash@example.com'

        self.spy_on(SandboxAuthBackend.update_name)

        form.save()
        self.assertTrue(SandboxAuthBackend.update_name.called)

    def test_update_email_auth_backend(self):
        """Testing ProfileForm.save with error in auth_backend.update_email"""
        form = ProfileForm(page=None, request=self.request, user=self.user)
        form.cleaned_data = {
            'first_name': 'Barry',
            'last_name': 'Allen',
            'email': 'flash@example.com',
            'profile_private': '',
        }
        self.user.first_name = 'Barry'
        self.user.last_name = 'Allen'

        self.spy_on(SandboxAuthBackend.update_email)

        form.save()
        self.assertTrue(SandboxAuthBackend.update_email.called)



"""Forms for searching Review Board."""

from __future__ import unicode_literals

from collections import OrderedDict

from django import forms
from django.contrib.auth.models import User
from django.utils import six
from django.utils.translation import ugettext_lazy as _
from haystack.forms import ModelSearchForm
from haystack.inputs import Raw
from haystack.query import SQ

from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.scmtools.models import Repository
from reviewboard.search.indexes import BaseSearchIndex


class RBSearchForm(ModelSearchForm):
    """The Review Board search form.

    This form is capable of searching for :py:class:`ReviewRequests
    <reviewboard.reviews.models.review_request.ReviewRequest>` and
    :py:class:`Users <django.contrib.auth.models.User>`.
    """

    FILTER_ALL = ''
    FILTER_REVIEW_REQUESTS = 'reviewrequests'
    FILTER_USERS = 'users'

    #: Available model filters.
    FILTER_TYPES = OrderedDict([
        (FILTER_ALL, {
            'models': [ReviewRequest, User],
            'name': _('All results'),
        }),
        (FILTER_REVIEW_REQUESTS, {
            'models': [ReviewRequest],
            'name': _('Review Requests'),
        }),
        (FILTER_USERS, {
            'models': [User],
            'name': _('Users'),
        }),
    ])

    model_filter = forms.MultipleChoiceField(
        choices=(
            (filter_id, filter_type['name'])
            for filter_id, filter_type in six.iteritems(FILTER_TYPES)
        ),
        required=False,
    )

    id = forms.IntegerField(required=False)

    def __init__(self, user=None, local_site=None, **kwargs):
        """Initialize the search form.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the search.

                Results will be limited to those visible to the user.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the search is being performed on.

                Results will be limited to those on the LocalSite.

            **kwargs (dict):
                Additional keyword arguments to forward to the parent form.
        """
        super(RBSearchForm, self).__init__(**kwargs)

        self.user = user
        self.local_site = local_site

    def clean_q(self):
        """Clean the ``q`` field.

        The field will be stripped of leading and trailing whitespace.

        Returns:
            unicode:
            The stripped query.
        """
        return self.cleaned_data['q'].strip()

    def clean_model_filter(self):
        """Clean the ``model_filter`` field.

        If no filter is provided, the default (all models) will be used.

        Returns:
            list of unicode:
            The cleaned ``filter`` field.
        """
        return self.cleaned_data['model_filter'] or ['']

    def search(self):
        """Perform a search.

        Returns:
            haystack.query.SearchQuerySet:
            The search results.
        """
        if not self.is_valid():
            return self.no_query_found()

        user = self.user
        q = self.cleaned_data['q']
        id_q = self.cleaned_data.get('id')
        model_filters = set()

        for filter_type in self.cleaned_data.get('model_filter', ['']):
            model_filters.update(self.FILTER_TYPES[filter_type]['models'])

        model_filters = list(model_filters)

        sqs = (
            self.searchqueryset
            .filter(content=Raw(q))
            .models(*model_filters)
        )

        if id_q:
            sqs = sqs.filter_or(SQ(id=q))

        if self.local_site:
            local_site_id = self.local_site.pk
        else:
            local_site_id = BaseSearchIndex.NO_LOCAL_SITE_ID

        sqs = sqs.filter_and(local_sites__contains=local_site_id)

        # Filter out any private review requests the user doesn't have
        # access to.
        if not user.is_superuser:
            private_sq = (SQ(django_ct='reviews.reviewrequest') &
                          SQ(private=True))

            if user.is_authenticated():
                # We're going to build a series of queries that mimic the
                # accessibility checks we have internally, based on the access
                # permissions the user currently has, and the IDs listed in
                # the indexed review request.
                #
                # This must always be kept in sync with
                # ReviewRequestManager._query.
                #
                # Note that we are not performing Local Site checks here,
                # because we're already filtering by Local Sites.

                # Make sure they have access to the repository, if any.
                accessible_repo_ids = list(Repository.objects.accessible_ids(
                    user,
                    visible_only=False,
                    local_site=self.local_site,
                ))

                accessible_group_ids = Group.objects.accessible_ids(
                    user,
                    visible_only=False,
                )

                repository_sq = SQ(
                    private_repository_id__in=[0] + accessible_repo_ids
                )

                # Next, build a query to see if the review request targets any
                # invite-only groups the user is a member of.
                target_groups_sq = SQ(private_target_groups__contains=0)

                for pk in accessible_group_ids:
                    target_groups_sq |= SQ(private_target_groups__contains=pk)

                # Build a query to see if the user is explicitly listed
                # in the list of reviewers.
                target_users_sq = SQ(target_users__contains=user.pk)

                # And, of course, the owner of the review request can see it.
                #
                # With that, we'll put the whole query together, in the order
                # matching ReviewRequest.is_accessible_by.
                private_sq &= ~(SQ(username=user.username) |
                                (repository_sq &
                                 (target_users_sq | target_groups_sq)))

            sqs = sqs.exclude(private_sq)

        return sqs.order_by('-last_updated')



"""Unit tests for reviewboard.reviews.views.RootView."""

from __future__ import unicode_literals

from djblets.siteconfig.models import SiteConfiguration
from djblets.testing.decorators import add_fixtures

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class RootViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.RootView."""

    fixtures = ['test_users']

    def test_with_anonymous_with_private_access(self):
        """Testing RootView with anonymous user with anonymous access not
        allowed
        """
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('auth_require_sitewide_login', True)
        siteconfig.save()

        response = self.client.get(local_site_reverse('root'))

        self.assertEqual(response['Location'],
                         'http://testserver/account/login/?next=/')

    def test_with_anonymous_with_public_access(self):
        """Testing RootView with anonymous user with anonymous access allowed
        """
        response = self.client.get(local_site_reverse('root'))

        self.assertEqual(response['Location'], 'http://testserver/r/')

    def test_with_logged_in(self):
        """Testing RootView with authenticated user"""
        self.assertTrue(self.client.login(username='doc', password='doc'))

        response = self.client.get(local_site_reverse('root'))

        self.assertEqual(response['Location'], 'http://testserver/dashboard/')

    @add_fixtures(['test_site'])
    def test_with_anonymous_with_local_site_private(self):
        """Testing RootView with anonymous user with private Local Site"""
        response = self.client.get(
            local_site_reverse('root', local_site_name=self.local_site_name))

        self.assertEqual(response['Location'],
                         'http://testserver/account/login/?next=/s/%s/'
                         % self.local_site_name)

    @add_fixtures(['test_site'])
    def test_with_anonymous_with_local_site_public(self):
        """Testing RootView with anonymous user with public Local Site"""
        local_site = self.get_local_site(name=self.local_site_name)
        local_site.public = True
        local_site.save()

        response = self.client.get(local_site_reverse('root',
                                                      local_site=local_site))

        self.assertEqual(response['Location'],
                         'http://testserver/s/%s/r/' % self.local_site_name)

    @add_fixtures(['test_site'])
    def test_with_logged_in_with_local_site(self):
        """Testing RootView with authenticated user with Local Site"""
        self.assertTrue(self.client.login(username='doc', password='doc'))

        response = self.client.get(
            local_site_reverse('root', local_site_name=self.local_site_name))

        self.assertEqual(response['Location'],
                         'http://testserver/s/%s/dashboard/'
                         % self.local_site_name)



"""NIS authentication backend."""

from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import NISSettingsForm


class NISBackend(BaseAuthBackend):
    """Authenticate against a user on an NIS server."""

    backend_id = 'nis'
    name = _('NIS')
    settings_form = NISSettingsForm
    login_instructions = \
        _('Use your standard NIS username and password.')

    def authenticate(self, username, password, **kwargs):
        """Authenticate the user.

        This will authenticate the username and return the appropriate User
        object, or None.
        """
        import crypt
        import nis

        username = username.strip()

        try:
            passwd = nis.match(username, 'passwd').split(':')
            original_crypted = passwd[1]
            new_crypted = crypt.crypt(password, original_crypted)

            if original_crypted == new_crypted:
                return self.get_or_create_user(username, None, passwd)
        except nis.error:
            # FIXME I'm not sure under what situations this would fail (maybe
            # if their NIS server is down), but it'd be nice to inform the
            # user.
            pass

        return None

    def get_or_create_user(self, username, request, passwd=None):
        """Get an existing user, or create one if it does not exist."""
        import nis

        username = username.strip()

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            try:
                if not passwd:
                    passwd = nis.match(username, 'passwd').split(':')

                names = passwd[4].split(',')[0].split(' ', 1)
                first_name = names[0]
                last_name = None
                if len(names) > 1:
                    last_name = names[1]

                email = '%s@%s' % (username, settings.NIS_EMAIL_DOMAIN)

                user = User(username=username,
                            password='',
                            first_name=first_name,
                            last_name=last_name or '',
                            email=email)
                user.is_staff = False
                user.is_superuser = False
                user.set_unusable_password()
                user.save()
            except nis.error:
                pass
        return user



from __future__ import unicode_literals

import logging
import os
import zlib
from collections import OrderedDict

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import FileSystemStorage
from django.utils import six
from django.utils.six.moves.urllib.error import HTTPError
from django.utils.six.moves.urllib.request import urlopen
from django.utils.translation import ngettext
from django.utils.translation import ugettext_lazy as _

from reviewboard.admin.server import build_server_url


_security_checks = OrderedDict()


class BaseSecurityCheck(object):
    """Base class for a security check."""

    name = None
    desc = None
    fix_info = None

    def setUp(self):
        """Set up the security check."""
        pass

    def execute(self):
        """Execute the security check.

        This must be implemented by subclasses.

        Returns:
            tuple:
            A tuple of ``(success, error_message)``.
        """
        raise NotImplementedError

    def tearDown(self):
        """Tear down the security check."""
        pass


class BaseExecutableFileCheck(BaseSecurityCheck):
    """Base class for a security check involving uploaded files.

    This handles registering files to check, storing them, and accessing
    them in order to determine whether a server-side or client-side
    vulnerability exists in the configuration around uploaded media files.
    """

    FILENAME_PREFIX = 'exec_security_check'

    def __init__(self):
        """Initialize the security check."""
        self.storage = FileSystemStorage(
            location=os.path.join(settings.MEDIA_ROOT, 'uploaded', 'files'))
        self.directory = '%suploaded/files/' % settings.MEDIA_URL

        self.file_checks = []

    def setUp(self):
        """Set up the security check.

        This will go through the various file extensions that we intend to
        check and create file attachments with the given content.
        """
        if self._using_default_storage():
            for i, file_check in enumerate(self.file_checks):
                extensions_list, content = file_check
                bad_extensions = set()

                for ext in extensions_list:
                    try:
                        self.storage.save('%s%s' % (self.FILENAME_PREFIX, ext),
                                          ContentFile(content))
                    except OSError:
                        # Some web server configurations prevent even saving
                        # files with certain extensions. In this case, things
                        # will definitely succeed.
                        bad_extensions.add(ext)

                # Filter out any extensions that we failed to save, because we
                # don't need to check that they downloaded properly.
                extensions_list = [
                    ext
                    for ext in extensions_list
                    if ext not in bad_extensions
                ]
                self.file_checks[i] = (extensions_list, content)

    def tearDown(self):
        """Tear down the security check.

        This will delete all of the files created in ``setUp``.
        """
        if self._using_default_storage():
            for extensions_list, content in self.file_checks:
                for ext in extensions_list:
                    self.storage.delete('%s%s' % (self.FILENAME_PREFIX, ext))

    def execute(self):
        """Execute the security check.

        This will download each file that was created in ``setUp`` and check
        that the content matches what we expect.

        Returns:
            tuple:
            A tuple of ``(success, error_message)``.
        """
        failed_exts = []

        if not self._using_default_storage():
            return True, None

        for extensions_list, content in self.file_checks:
            for ext in extensions_list:
                try:
                    filename = '%s%s' % (self.FILENAME_PREFIX, ext)
                    url = build_server_url('%s/%s'
                                           % (self.directory.rstrip('/'),
                                              filename))

                    if not self.check_file(filename, url):
                        failed_exts.append(ext)
                except Exception as e:
                    return (False, _('Uncaught exception during test: %s') % e)

        if failed_exts:
            return (
                False,
                ngettext(
                    'The web server incorrectly executed this file type: %s',
                    'The web server incorrectly executed these file types: %s',
                    len(failed_exts)
                ) % ', '.join(failed_exts)
            )

        return True, None

    def _using_default_storage(self):
        """Return whether the server is using the built-in file storage.

        Returns:
            bool:
            ``True`` if the default (local) storage backend is being used.
            ``False`` if a different backend is being used.
        """
        return (settings.DEFAULT_FILE_STORAGE ==
                'django.core.files.storage.FileSystemStorage')


class ServerExecutableFileCheck(BaseExecutableFileCheck):
    """Check that uploaded files aren't executed server-side.

    Web servers like to run code in files named things like .php or .shtml.
    This check makes sure that user-uploaded files do not get executed when
    loading them via their URL.
    """

    name = _("Checking that uploaded files won't be executed by the server")
    desc = _('A misconfiguration in the web server can cause files attached '
             'to review requests to be executed as code. The file types '
             'checked in this test are: .html, .htm, .shtml, .php, .php3, '
             '.php4, .php5, .phps, .asp, .pl, .py, .fcgi, .cgi, .phtml, '
             '.phtm, .pht, .jsp, .sh, and .rb.')
    fix_info = _('For instructions on how to fix this problem, please visit '
                 '<a href="http://support.beanbaginc.com/support/solutions/'
                 'articles/110173-securing-file-attachments">'
                 'http://support.beanbaginc.com/support/solutions/articles/'
                 '110173-securing-file-attachments</a>')

    def setUp(self):
        """Set up the security check."""
        self.file_checks = [
            (
                ['.php', '.php3', '.php4', '.php5', '.phps', '.phtml',
                 '.phtm'],
                '<?php echo "Hello, World!"; ?>'
            ),
            (
                ['.pl', '.py'],
                'print "Hello, World!"'
            ),
            (
                ['.html', '.htm', '.shtml', '.pht'],
                ('<HTML>\n'
                 '<HEAD>\n'
                 '<TITLE>Hello, world!</TITLE>\n'
                 '</HEAD>\n'
                 '<BODY>\n'
                 '<H1>Hello, world!</H1>\n'
                 '<!--#echo var="LAST_MODIFIED" -->\n'
                 '<!--#exec cmd="echo HI!" -->\n'
                 '</BODY>\n'
                 '</HTML>')
            ),
            (
                ['.jsp'],
                '<%= new String("Hello!") %>'
            ),
            (
                ['.asp'],
                '<%="Hello World!"%>'
            ),
            (
                ['.fcgi', '.cgi', '.sh'],
                ('#!/bin/sh\n'
                 'echo "Hello World!"')
            ),
            (
                ['.rb'],
                'puts "Hello world!"'
            )
        ]

        super(ServerExecutableFileCheck, self).setUp()

    def check_file(self, filename, url):
        """Download a file and compare the resulting response to the file.

        This makes sure that when we fetch a file via its URL, the returned
        contents are identical to the file contents. This returns True if the
        file contents match, and False otherwise.

        Args:
            filename (unicode):
                The name of the file.

            url (unicode):
                The URL of the file.

        Returns:
            bool:
            ``True`` if the file could be downloaded (or a HTTP 403 was hit)
            and the contents matched the expected value.

            ``False`` if the download failed for some reason or the contents
            didn't match expectations.
        """
        try:
            data = urlopen(url).read()
        except HTTPError as e:
            # An HTTP 403 is also an acceptable response
            if e.code == 403:
                return True
            else:
                raise e

        with self.storage.open(filename, 'r') as f:
            return data == f.read()


class BrowserExecutableFileCheck(BaseExecutableFileCheck):
    """Check that uploaded files won't be executed client-side.

    Some file types (like SVGs and HTML files) that can be viewed inline in the
    browser once downloaded are also capable of running JavaScript. When this
    happens, those scripts may have access to the same cookies and sessions
    allowed by Review Board itself (if Review Board is hosting the files).
    These security checks ensure that the server is set up to force these
    files to download, rather than allow inline viewing.
    """

    name = _("Checking downloaded files aren't executable by the browser")
    desc = _(
        'Certain file types (such as SVG or HTML files) can contain '
        'embedded scripts that would be executed by your browser when viewed. '
        'We recommend forcing all uploaded media files to download when '
        'directly accessed in a browser. This applies only to files being '
        'served by Review Board, and not from a CDN.'
    )
    fix_info = _('For instructions on how to fix this problem, please visit '
                 '<a href="http://support.beanbaginc.com/support/solutions/'
                 'articles/110173-securing-file-attachments">'
                 'http://support.beanbaginc.com/support/solutions/articles/'
                 '110173-securing-file-attachments</a>')

    def setUp(self):
        """Set up the security check."""
        self.file_checks = [
            (
                ['.htm', '.html', '.shtml', '.stm', '.shtm'],
                '<html><body><script>alert("Hello!")</script></body></html>',
            ),
            (
                ['.svg'],
                '<svg xmlns="http://www.w3.org/2000/svg">'
                ' <script>alert("Hello!")</script>'
                '</svg>',
            ),
            (
                ['.svgz'],
                zlib.compress(
                    '<svg xmlns="http://www.w3.org/2000/svg">'
                    ' <script>alert("Hello!")</script>'
                    '</svg>',
                ),
            ),
        ]

        super(BrowserExecutableFileCheck, self).setUp()

    def check_file(self, filename, url):
        """Download a file and check for expected headers.

        This makes sure that when we fetch a file via its URL, the returned
        file's headers would instruct the browser to force a download, rather
        than view the contents inline.

        Args:
            filename (unicode):
                The name of the file.

            url (unicode):
                The URL of the file.

        Returns:
            bool:
            ``True`` if the file could be downloaded and the headers
            contain ``Content-Disposition: attachment``.

            ``False`` if the download failed for some reason or the header
            was not what we expected.
        """
        # Exceptions coming from this will be caught higher up.
        headers = urlopen(url).info()

        return headers.get('Content-Disposition', '').startswith('attachment')


class AllowedHostsCheck(BaseSecurityCheck):
    """Check that the ALLOWED_HOSTS setting is configured.

    In order to prevent URL inejections, Django requires that ALLOWED_HOSTS be
    configured with a list of hostnames for which Review Board will answer.
    People upgrading from previous versions will have this set to a wildcard.
    """

    name = _('Checking ALLOWED_HOSTS setting')
    desc = _('ALLOWED_HOSTS is a list containing the host/domain names that '
             'Review Board will consider valid for this server to serve. '
             'This is a security measure to prevent an attacker from '
             'poisoning cache and password reset e-mails with links to '
             'malicious hosts by submitting requests with a fake HTTP Host '
             'header, which is possible even under many seemingly-safe web '
             'server configurations.')
    fix_info = _("To fix this, edit the settings_local.py in the site's conf "
                 "directory and add a line like this with your site's URL: "
                 "<pre>ALLOWED_HOSTS = ['example.com']</pre>")

    def execute(self):
        """Execute the security check.

        This checks the value of the ALLOWED_HOSTS setting to make sure that it
        contains one or more hostnames.
        """
        result = True
        error_msg = ''

        if len(settings.ALLOWED_HOSTS) < 1:
            result = False
            error_msg = _('ALLOWED_HOSTS is empty.')

        if '*' in settings.ALLOWED_HOSTS:
            result = False
            error_msg = _("ALLOWED_HOSTS contains '*', which means that the "
                          "server will respond to any host.")

        return result, error_msg


class SecurityCheckRunner(object):
    """This is a runner to execute the security checks defined above.

    In order for a check to be run in this runner it needs to be added
    to the _security_checks list.

    The information that comes back from a single check is the following:
    - name: User-friendly name used to describe the check.
    - desc: A more detailed description to provide information about the check.
    - result: True if the check passed, or False if it failed or there was
              an excetion during its execution.
    - error_msg: A description of what failed. This will be blank if the test
                 passes.
    - fix_info: Instructions containing what a user should do if a check fails.
    """

    def __init__(self):
        """Initialize the security check runner."""
        pass

    def run(self):
        """Run all security checks and return the results."""
        all_test_results = []
        checks = get_security_checks()

        for name, cls in six.iteritems(checks):
            check = cls()

            check.setUp()
            current_test_result, error_msg = check.execute()
            check.tearDown()

            all_test_results.append({
                'name': check.name,
                'desc': check.desc,
                'result': current_test_result,
                'error_msg': error_msg,
                'fix_info': check.fix_info,
            })

        return all_test_results


def _populate_security_checks():
    """Populate a list of existing security checks."""
    if not _security_checks:
        _security_checks['hosts_check'] = AllowedHostsCheck
        _security_checks['server_executable_check'] = ServerExecutableFileCheck
        _security_checks['browser_executable_file_check'] = \
            BrowserExecutableFileCheck


def get_security_checks():
    """Return the list of security checks."""
    _populate_security_checks()

    return _security_checks


def register_security_check(name, cls):
    """Register a custom security check."""
    _populate_security_checks()

    if name in _security_checks:
        raise KeyError('"%s" is already a registered security check' % name)

    _security_checks[name] = cls


def unregister_security_check(name):
    """Unregister a previously registered security check."""
    _populate_security_checks()

    try:
        del _security_checks[name]
    except KeyError:
        logging.error('Failed to unregister unknown security check "%s"' %
                      name)
        raise KeyError('"%s" is not a registered security check' % name)



"""Condition support for Local Sites."""

from __future__ import unicode_literals


class LocalSiteModelChoiceMixin(object):
    """Mixin to restrict model choices to those on a Local Site.

    This will ensure that any queries are bound to a
    :py:class:`~reviewboard.site.models.LocalSite`, if provided to the
    condition field widget through
    :py:attr:`~djblets.forms.widgets.ConditionsWidget.choice_kwargs`.
    """

    #: The field on the model pointing to a LocalSite.
    local_site_attr = 'local_site'

    def get_queryset(self):
        """Return a queryset for the choice.

        The queryset will take into account a LocalSite, if one is provided
        for the conditions.

        Returns:
            django.db.models.query.QuerySet:
            The queryset for the choice.
        """
        queryset = super(LocalSiteModelChoiceMixin, self).get_queryset()
        local_site = self.extra_state.get('local_site')

        if local_site:
            queryset = queryset.filter(**{
                self.local_site_attr: local_site,
            })

        return queryset



# coding=utf-8
from __future__ import unicode_literals

import os

import nose

from reviewboard.scmtools.core import PRE_CREATION, Revision
from reviewboard.scmtools.errors import SCMError, FileNotFoundError
from reviewboard.scmtools.hg import HgDiffParser, HgGitDiffParser
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.testing import online_only


class MercurialTests(SCMTestCase):
    """Unit tests for mercurial."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(MercurialTests, self).setUp()

        hg_repo_path = os.path.join(os.path.dirname(__file__),
                                    '..', 'testdata', 'hg_repo')
        self.repository = Repository(name='Test HG',
                                     path=hg_repo_path,
                                     tool=Tool.objects.get(name='Mercurial'))

        try:
            self.tool = self.repository.get_scmtool()
        except ImportError:
            raise nose.SkipTest('Hg is not installed')

    def _first_file_in_diff(self, diff):
        return self.tool.get_parser(diff).parse()[0]

    def test_ssh_disallowed(self):
        """Testing HgTool does not allow SSH URLs"""
        with self.assertRaises(SCMError):
            self.tool.check_repository('ssh://foo')

    def test_git_parser_selection_with_header(self):
        """Testing HgTool returns the git parser when a header is present"""
        diffContents = (b'# HG changeset patch\n'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff --git a/emptyfile b/emptyfile\n'
                        b'new file mode 100644\n')

        parser = self.tool.get_parser(diffContents)
        self.assertEqual(type(parser), HgGitDiffParser)

    def test_hg_parser_selection_with_header(self):
        """Testing HgTool returns the hg parser when a header is present"""
        diffContents = (b'# HG changeset patch'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff -r 9d3f4147f294 -r 6187592a72d7 new.py\n'
                        b'--- /dev/null   Thu Jan 01 00:00:00 1970 +0000\n'
                        b'+++ b/new.py  Tue Apr 21 12:20:05 2015 -0400\n')

        parser = self.tool.get_parser(diffContents)
        self.assertEqual(type(parser), HgDiffParser)

    def test_git_parser_sets_commit_ids(self):
        """Testing HgGitDiffParser sets the parser commit ids"""
        diffContents = (b'# HG changeset patch\n'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff --git a/emptyfile b/emptyfile\n'
                        b'new file mode 100644\n')

        parser = self.tool.get_parser(diffContents)
        parser.parse()
        self.assertEqual(parser.new_commit_id, b'6187592a72d7')
        self.assertEqual(parser.base_commit_id, b'9d3f4147f294')

    def test_patch_creates_new_file(self):
        """Testing HgTool with a patch that creates a new file"""
        self.assertEqual(
            PRE_CREATION,
            self.tool.parse_diff_revision('/dev/null', 'bf544ea505f8')[1])

    def test_diff_parser_new_file(self):
        """Testing HgDiffParser with a diff that creates a new file"""
        diffContents = (b'diff -r bf544ea505f8 readme\n'
                        b'--- /dev/null\n'
                        b'+++ b/readme\n')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origFile, 'readme')

    def test_diff_parser_with_added_empty_file(self):
        """Testing HgDiffParser with a diff with an added empty file"""
        diff = (b'diff -r 356a6127ef19 -r 4960455a8e88 empty\n'
                b'--- /dev/null\n'
                b'+++ b/empty\n')

        file = self._first_file_in_diff(diff)
        self.assertEqual(file.origInfo, PRE_CREATION)
        self.assertEqual(file.origFile, 'empty')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'empty')
        self.assertFalse(file.binary)
        self.assertFalse(file.deleted)
        self.assertEqual(file.insert_count, 0)
        self.assertEqual(file.delete_count, 0)

    def test_diff_parser_with_deleted_empty_file(self):
        """Testing HgDiffParser with a diff with a deleted empty file"""
        diff = (b'diff -r 356a6127ef19 -r 4960455a8e88 empty\n'
                b'--- a/empty\n'
                b'+++ /dev/null\n')

        file = self._first_file_in_diff(diff)
        self.assertEqual(file.origInfo, '356a6127ef19')
        self.assertEqual(file.origFile, 'empty')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'empty')
        self.assertFalse(file.binary)
        self.assertTrue(file.deleted)
        self.assertEqual(file.insert_count, 0)
        self.assertEqual(file.delete_count, 0)

    def test_diff_parser_uncommitted(self):
        """Testing HgDiffParser with a diff with an uncommitted change"""
        diffContents = (b'diff -r bf544ea505f8 readme\n'
                        b'--- a/readme\n'
                        b'+++ b/readme\n')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, 'bf544ea505f8')
        self.assertEqual(file.origFile, 'readme')
        self.assertEqual(file.newInfo, 'Uncommitted')
        self.assertEqual(file.newFile, 'readme')

    def test_diff_parser_committed(self):
        """Testing HgDiffParser with a diff between committed revisions"""
        diffContents = (b'diff -r 356a6127ef19 -r 4960455a8e88 readme\n'
                        b'--- a/readme\n'
                        b'+++ b/readme\n')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, '356a6127ef19')
        self.assertEqual(file.origFile, 'readme')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'readme')

    def test_diff_parser_with_preamble_junk(self):
        """Testing HgDiffParser with a diff that contains non-diff junk test
        as a preamble
        """
        diffContents = (b'changeset:   60:3613c58ad1d5\n'
                        b'user:        Michael Rowe <mrowe@mojain.com>\n'
                        b'date:        Fri Jul 27 11:44:37 2007 +1000\n'
                        b'files:       readme\n'
                        b'description:\n'
                        b'Update the readme file\n'
                        b'\n'
                        b'\n'
                        b'diff -r 356a6127ef19 -r 4960455a8e88 readme\n'
                        b'--- a/readme\n'
                        b'+++ b/readme\n')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, '356a6127ef19')
        self.assertEqual(file.origFile, 'readme')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'readme')

    def test_git_diff_parsing(self):
        """Testing HgDiffParser git diff support"""
        diffContents = (b'# Node ID 4960455a8e88\n'
                        b'# Parent bf544ea505f8\n'
                        b'diff --git a/path/to file/readme.txt '
                        b'b/new/path to/readme.txt\n'
                        b'rename from path/to file/readme.txt\n'
                        b'rename to new/path to/readme.txt\n'
                        b'--- a/path/to file/readme.txt\n'
                        b'+++ b/new/path to/readme.txt\n')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, 'bf544ea505f8')
        self.assertEqual(file.origFile, 'path/to file/readme.txt')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'new/path to/readme.txt')

    def test_diff_parser_unicode(self):
        """Testing HgDiffParser with unicode characters"""

        diffContents = ('diff -r bf544ea505f8 réadme\n'
                        '--- a/réadme\n'
                        '+++ b/réadme\n').encode('utf-8')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, 'bf544ea505f8')
        self.assertEqual(file.origFile, 'réadme')
        self.assertEqual(file.newInfo, 'Uncommitted')
        self.assertEqual(file.newFile, 'réadme')

    def test_git_diff_parsing_unicode(self):
        """Testing HgDiffParser git diff with unicode characters"""
        diffContents = ('# Node ID 4960455a8e88\n'
                        '# Parent bf544ea505f8\n'
                        'diff --git a/path/to file/réadme.txt '
                        'b/new/path to/réadme.txt\n'
                        'rename from path/to file/réadme.txt\n'
                        'rename to new/path to/réadme.txt\n'
                        '--- a/path/to file/réadme.txt\n'
                        '+++ b/new/path to/réadme.txt\n').encode('utf-8')

        file = self._first_file_in_diff(diffContents)
        self.assertEqual(file.origInfo, 'bf544ea505f8')
        self.assertEqual(file.origFile, 'path/to file/réadme.txt')
        self.assertEqual(file.newInfo, '4960455a8e88')
        self.assertEqual(file.newFile, 'new/path to/réadme.txt')

    def test_revision_parsing(self):
        """Testing HgDiffParser revision number parsing"""
        self.assertEqual(
            self.tool.parse_diff_revision('doc/readme', 'bf544ea505f8'),
            ('doc/readme', 'bf544ea505f8'))

        self.assertEqual(
            self.tool.parse_diff_revision('/dev/null', 'bf544ea505f8'),
            ('/dev/null', PRE_CREATION))

        # TODO think of a meaningful thing to test here...
        # self.assertRaises(SCMException,
        #                  lambda: self.tool.parse_diff_revision('', 'hello'))

    def test_get_branches(self):
        """Testing list of branches in HgClient.get_change"""
        value = self.tool.get_branches()
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

        self.assertEqual(value[0].id, 'default')
        self.assertEqual(value[0].commit,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].default, True)

    def test_get_change(self):
        """Testing raw diff of HgClient.get_change"""
        self.assertRaises(SCMError, lambda: self.tool.get_change('dummy'))

        value = self.tool.get_change('0')
        self.assertNotIn('goodbye', value.diff)
        self.assertEqual(value.id, 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        value = self.tool.get_change('1')
        self.assertIn('goodbye', value.diff)
        self.assertEqual(value.id, '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')

    def test_get_commits(self):
        """Testing commit objects in HgClient.get_commits"""
        value = self.tool.get_commits()
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 2)

        self.assertEqual(value[0].id,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].message, 'second')
        self.assertEqual(value[0].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[0].date, '2007-08-07T17:12:23')
        self.assertEqual(value[0].parent,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[0].base_commit_id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')

        self.assertEqual(value[1].id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[1].message, 'first')
        self.assertEqual(value[1].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[1].date, '2007-08-07T17:11:57')
        self.assertEqual(value[1].parent,
                         '0000000000000000000000000000000000000000')
        self.assertEqual(value[1].base_commit_id,
                         '0000000000000000000000000000000000000000')

        self.assertRaisesRegexp(SCMError, 'Cannot load commits: ',
                                lambda: self.tool.get_commits(branch='x'))

        rev = 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786'
        value = self.tool.get_commits(start=rev)
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

    def test_get_commits_with_non_utc_server_timezone(self):
        """Testing commit objects in HgClient.get_commits with
        settings.TIME_ZONE != UTC
        """
        old_tz = os.environ[b'TZ']
        os.environ[b'TZ'] = b'US/Pacific'

        try:
            value = self.tool.get_commits()
        finally:
            os.environ[b'TZ'] = old_tz

        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 2)

        self.assertEqual(value[0].id,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].message, 'second')
        self.assertEqual(value[0].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[0].date, '2007-08-07T17:12:23')
        self.assertEqual(value[0].parent,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[0].base_commit_id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')

        self.assertEqual(value[1].id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[1].message, 'first')
        self.assertEqual(value[1].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[1].date, '2007-08-07T17:11:57')
        self.assertEqual(value[1].parent,
                         '0000000000000000000000000000000000000000')
        self.assertEqual(value[1].base_commit_id,
                         '0000000000000000000000000000000000000000')

        self.assertRaisesRegexp(SCMError, 'Cannot load commits: ',
                                lambda: self.tool.get_commits(branch='x'))

        rev = 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786'
        value = self.tool.get_commits(start=rev)
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

    def test_get_file(self):
        """Testing HgTool.get_file"""
        rev = Revision('661e5dd3c493')
        file = 'doc/readme'

        value = self.tool.get_file(file, rev)
        self.assertTrue(isinstance(value, bytes))
        self.assertEqual(value, b'Hello\n\ngoodbye\n')

        self.assertTrue(self.tool.file_exists('doc/readme', rev))
        self.assertTrue(not self.tool.file_exists('doc/readme2', rev))

        self.assertRaises(FileNotFoundError, lambda: self.tool.get_file(''))

        self.assertRaises(FileNotFoundError,
                          lambda: self.tool.get_file('hello', PRE_CREATION))

    def test_get_file_base_commit_id_override(self):
        """Testing base_commit_id overrides revision in HgTool.get_file"""
        base_commit_id = Revision('661e5dd3c493')
        bogus_rev = Revision('bogusrevision')
        file = 'doc/readme'

        value = self.tool.get_file(file, bogus_rev,
                                   base_commit_id=base_commit_id)
        self.assertTrue(isinstance(value, bytes))
        self.assertEqual(value, b'Hello\n\ngoodbye\n')

        self.assertTrue(self.tool.file_exists(
            'doc/readme',
            bogus_rev,
            base_commit_id=base_commit_id))
        self.assertTrue(not self.tool.file_exists(
            'doc/readme2',
            bogus_rev,
            base_commit_id=base_commit_id))

    def test_interface(self):
        """Testing basic HgTool API"""
        self.assertTrue(self.tool.diffs_use_absolute_paths)

        self.assertRaises(NotImplementedError,
                          lambda: self.tool.get_changeset(1))

    @online_only
    def test_https_repo(self):
        """Testing HgTool.file_exists with an HTTPS-based repository"""
        repo = Repository(name='Test HG2',
                          path='https://bitbucket.org/pypy/pypy',
                          tool=Tool.objects.get(name='Mercurial'))
        tool = repo.get_scmtool()

        rev = Revision('877cf1960916')

        self.assertTrue(tool.file_exists('TODO.rst', rev))
        self.assertTrue(not tool.file_exists('TODO.rstNotFound', rev))



"""Unit tests for reviewboard.accounts.forms.pages.PrivacyForm."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.test.client import RequestFactory

from reviewboard.accounts.forms.pages import PrivacyForm
from reviewboard.accounts.pages import PrivacyPage
from reviewboard.accounts.views import MyAccountView
from reviewboard.testing import TestCase


class PrivacyFormTests(TestCase):
    """Unit tests for reviewboard.accounts.forms.pages.PrivacyForm."""

    def setUp(self):
        super(PrivacyFormTests, self).setUp()

        self.user = User.objects.create(username='test-user')

        self.request = RequestFactory().get('/account/preferences/')
        self.request.user = self.user

        self.page = PrivacyPage(config_view=MyAccountView(),
                                request=self.request,
                                user=self.user)

    def test_init_with_privacy_enable_user_consent_true(self):
        """Testing PrivacyForm with privacy_enable_user_consent=True"""
        with self.siteconfig_settings({'privacy_enable_user_consent': True}):
            form = PrivacyForm(page=self.page,
                               request=self.request,
                               user=self.user)
            self.assertIn('consent', form.fields)
            self.assertEqual(form.save_label, 'Save')

    def test_init_with_privacy_enable_user_consent_false(self):
        """Testing PrivacyForm with privacy_enable_user_consent=False"""
        with self.siteconfig_settings({'privacy_enable_user_consent': False}):
            form = PrivacyForm(page=self.page,
                               request=self.request,
                               user=self.user)
            self.assertNotIn('consent', form.fields)
            self.assertIsNone(form.save_label)

    def test_is_visible_with_no_privacy(self):
        """Testing PrivacyForm.is_visible with no privacy details"""
        settings = {
            'privacy_enable_user_consent': False,
            'privacy_info_html': '',
        }

        with self.siteconfig_settings(settings):
            form = PrivacyForm(page=self.page,
                               request=self.request,
                               user=self.user)
            self.assertFalse(form.is_visible())

    def test_is_visible_with_consent(self):
        """Testing PrivacyForm.is_visible with consent option enabled"""
        settings = {
            'privacy_enable_user_consent': True,
            'privacy_info_html': '',
        }

        with self.siteconfig_settings(settings):
            form = PrivacyForm(page=self.page,
                               request=self.request,
                               user=self.user)
            self.assertTrue(form.is_visible())

    def test_is_visible_with_privacy_info(self):
        """Testing PrivacyForm.is_visible with privacy_info_html set"""
        settings = {
            'privacy_enable_user_consent': False,
            'privacy_info_html': 'Test.',
        }

        with self.siteconfig_settings(settings):
            form = PrivacyForm(page=self.page,
                               request=self.request,
                               user=self.user)
            self.assertTrue(form.is_visible())



from __future__ import unicode_literals

from django.db.models.query import Q
from django.utils import six
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.resources.user import UserResource as DjbletsUserResource

from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.search import search_backend_registry
from reviewboard.search.forms import RBSearchForm
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)


class SearchResource(WebAPIResource, DjbletsUserResource):
    """
    Provides information on users, groups and review requests.

    This is the resource for the autocomplete widget for quick search. This
    resource helps filter for users, groups and review requests.
    """
    added_in = '1.6'

    name = 'search'
    singleton = True

    MIN_SUMMARY_LEN = 4

    def has_access_permissions(self, *args, **kwargs):
        """Return whether or not users have access to this resource.

        This resource is accessible to any users that have access to the API.

        Args:
            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            bool:
            Always ``True``.
        """
        return True

    @webapi_request_fields(
        optional={
            'q': {
                'type': six.text_type,
                'description': 'The text to search for.',
            },
            'displayname': {
                'type': bool,
                'description': 'This field is deprecated and ignored. It '
                               'will be removed in a future release of '
                               'Review Board.',
            },
            'fullname': {
                'type': bool,
                'description': 'Whether or not to include users whose full '
                               'name includes the search text.',
            },
            'id': {
                'type': int,
                'description': 'A specific review request ID to search for.',
            },
            'max_results': {
                'type': int,
                'description': 'The maximum number of results to return '
                               'for each type of matching object. By '
                               'default, this is 25. There is a hard limit '
                               'of 200.',
            },
        },
        allow_unknown=True,
    )
    @webapi_check_local_site
    @webapi_check_login_required
    def get(self, request, max_results=None, *args, **kwargs):
        """Returns information on users, groups and review requests.

        This is used by the autocomplete widget for quick search to get
        information on users, groups and review requests. This function returns
        users' first name, last name and username, groups' name and display
        name, and review requests' ID and summary.
        """
        max_results = min((max_results or 25), 200)

        try:
            # We have to keep the parameter named id for backwards
            # compatibility, but it would override the builtin of the same
            # name.
            kwargs['id_q'] = kwargs.pop('id')
        except KeyError:
            pass

        return 200, {
            self.name: {
                'users': self._search_users(
                    request=request,
                    max_results=max_results,
                    *args,
                    **kwargs
                ),
                'groups': self._search_groups(
                    request=request,
                    max_results=max_results,
                    *args,
                    **kwargs
                ),
                'review_requests': self._search_review_requests(
                    request=request,
                    max_results=max_results,
                    *args,
                    **kwargs
                )
            },
        }

    def _search_users(self, request, max_results, local_site=None,
                      fullname=None, q=None, id_q=None, *args, **kwargs):
        """Search for users and return the results.

        Args:
            request (django.http.HttpRequest):
                The current request.

            max_results (int):
                The maximum number of results to return.

            local_site (reviewboard.site.models.LocalSite, optional):
                The current local site.

            fullname (bool, optional):
                Whether or not to perform a search against the users' full
                names.

            q (unicode, optional):
                The search text.

            id_q (int, optional):
                An optional ID to search against user IDs.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            django.db.models.query.QuerySet or list:
            A query set for users matching the given arguments.
        """
        if (search_backend_registry.search_enabled and
            search_backend_registry.on_the_fly_indexing_enabled):
            # If search is enabled, we will use the index to perform the query.
            form = RBSearchForm(
                user=request.user,
                local_site=local_site,
                data={
                    'q': q,
                    'id_q': id_q,
                    'model_filter': [RBSearchForm.FILTER_USERS],
                }
            )

            results = []

            for result in form.search()[:max_results]:
                raw_user = {
                    'id': result.pk,
                    'username': result.username,
                    'url': result.url,
                }

                if not result.is_profile_private:
                    raw_user['fullname'] = result.full_name

                results.append(raw_user)

            return results

        # If search is disabled, we will fall back to using database queries.
        if local_site:
            users = local_site.users.filter(is_active=True)
        else:
            users = self.model.objects.filter(is_active=True)

        if q:
            parts = q.split(' ', 1)

            if len(parts) > 1:
                query = (
                    (Q(first_name__istartswith=parts[0]) &
                     Q(last_name__istartswith=parts[1])) |
                    (Q(first_name__istartswith=parts[1]) &
                     Q(last_name__istartswith=parts[0]))
                )

                if fullname:
                    query |= (Q(first_name__istartswith=q) |
                              Q(last_name__istartswith=q))

                query &= Q(profile__is_private=False)
            else:
                query = (Q(username__istartswith=q) |
                         (Q(profile__is_private=False) &
                          (Q(first_name__istartswith=q) |
                           Q(last_name__istartswith=q))))

            users = users.filter(query)

        return users[:max_results]

    def _search_groups(self, request, max_results, local_site=None, q=None,
                       *args, **kwargs):
        """Search for review groups and return the results.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            max_results (int):
                The maximum number of results to return.

            local_site (reviewboard.site.models.LocalSite, optional):
                The current local site.

            q (unicode, optional):
                The search text.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A query set for review groups matching the given arguments.
        """
        groups = Group.objects.accessible(request.user, local_site=local_site)

        if q:
            groups = groups.filter(
                Q(name__istartswith=q) |
                Q(display_name__istartswith=q)
            )

        # Group.objects.accessible only respects visible_only for
        # non-superusers. We add this here to make the behavior consistent.
        return groups.filter(visible=True)[:max_results]

    def _search_review_requests(self, request, max_results, local_site=None,
                                q=None, id_q=None, *args, **kwargs):
        """Search for a review request and return the results.

        If indexed search is enabled, this will use the search index. Otherwise
        it will query against the database.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The current local site.

            max_results (int):
                The maximum number of results to return.

            q (unicode, optional):
                The search text.

            id_q (int, optional):
                An optional ID to search against review request IDs.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            django.db.models.query.QuerySet or haystack.query.SearchQuerySet:
            A query for review requests matching the given arguments.
        """
        if (search_backend_registry.search_enabled and
            search_backend_registry.on_the_fly_indexing_enabled):
            # If search is enabled, we will use the index to perform the query.
            form = RBSearchForm(
                user=request.user,
                local_site=local_site,
                data={
                    'q': q,
                    'id': id_q,
                    'model_filter': [RBSearchForm.FILTER_REVIEW_REQUESTS],
                }
            )

            return [
                {
                    'id': result.review_request_id,
                    'public': True,  # Drafts are not indexed.
                    'summary': result.summary,
                }
                for result in form.search()[:max_results]
            ]

        # If search is disabled, we will fall back to using database queries.
        review_requests = ReviewRequest.objects.public(
            filter_private=True,
            user=request.user,
            local_site=local_site,
            status=None,
        )

        query = Q()

        if q:
            if local_site:
                query |= Q(local_id__istartswith=q)
            else:
                query |= Q(id__startswith=q)

            if len(q) >= self.MIN_SUMMARY_LEN:
                query |= Q(summary__istartswith=q)

        if id_q:
            if local_site:
                query |= Q(local_id__startswith=id_q)
            else:
                query |= Q(id__startswith=id_q)

        return review_requests.filter(query)[:max_results]


search_resource = SearchResource()
